"""."""

import time as _time
from threading import Event as _Event

import numpy as _np
from mathphys.functions import get_namedtuple as _get_namedtuple

from ..namesys import SiriusPVName as _PVName
from .dcct import DCCT
from .device import Device as _Device, DeviceSet as _DeviceSet
from .fpmosc import FPMOsc
from .rf import RFCav


class BunchbyBunch(_DeviceSet):
    """BunchbyBunch Device."""

    _devices = {
        'H': 'SI-Glob:DI-BbBProc-H',
        'V': 'SI-Glob:DI-BbBProc-V',
        'L': 'SI-Glob:DI-BbBProc-L',
    }
    DEVICES = _get_namedtuple('Devices', *zip(*_devices.items()))
    FBEDEV_PLANE = 'V'

    def __init__(self, devname, props2init='all'):
        """BunchbyBunch DeviceSet.

        Args:
            devname (str): choose one of the BbB planes in BunchbyBunch.DEVICES
            props2init (str, optional): 'all' to connect with all PVs or
                bool(props2init) == False to initialize without any
                connection. Defaults to 'all'.

        """
        devname = BunchbyBunch.process_device_name(devname)

        _isall = isinstance(props2init, str) and props2init.lower() == 'all'
        if not _isall and props2init:
            raise ValueError(
                "props2init must be 'all' or bool(props2init) == False"
            )

        self.fpmosc = FPMOsc(FPMOsc.DEVICES.SI)
        self.dcct = DCCT(DCCT.DEVICES.SI_13C4, props2init=props2init)
        self.rfcav_a = RFCav(RFCav.DEVICES.SIA, props2init=props2init)
        self.rfcav_b = RFCav(RFCav.DEVICES.SIB, props2init=props2init)
        self.info = SystemInfo(devname, props2init=props2init)
        self.timing = Timing(devname, props2init=props2init)
        self.sram = Acquisition(devname, acqtype='SRAM', props2init=props2init)
        self.bram = Acquisition(devname, acqtype='BRAM', props2init=props2init)
        self.single_bunch = SingleBunch(devname, props2init=props2init)
        self.phase_track = PhaseTracking(devname, props2init=props2init)
        self.coeffs = Coefficients(devname, props2init=props2init)
        self.feedback = Feedback(devname, props2init=props2init)
        self.drive0 = Drive(devname, drive_num=0, props2init=props2init)
        self.drive1 = Drive(devname, drive_num=1, props2init=props2init)
        self.drive2 = Drive(devname, drive_num=2, props2init=props2init)
        self.bunch_clean = BunchClean(devname, props2init=props2init)
        fbe_dev = devname.substitute(idx=self.FBEDEV_PLANE)
        self.fbe = FrontBackEnd(fbe_dev, props2init=props2init)
        devs = [
            self.info,
            self.timing,
            self.sram,
            self.bram,
            self.coeffs,
            self.feedback,
            self.drive0,
            self.drive1,
            self.drive2,
            self.bunch_clean,
            self.fbe,
            self.dcct,
            self.fpmosc,
            self.rfcav_a,
            self.rfcav_b,
            self.single_bunch,
            self.phase_track,
        ]

        if devname.endswith('-L'):
            self.pwr_amp1 = PwrAmpL(devname, num=0, props2init=props2init)
            self.pwr_amp2 = PwrAmpL(devname, num=1, props2init=props2init)
            devs.append(self.pwr_amp1)
            devs.append(self.pwr_amp2)
        else:
            self.pwr_amp = PwrAmpT(devname, props2init=props2init)
            devs.append(self.pwr_amp)

        super().__init__(devices=devs, devname=devname)

    @staticmethod
    def process_device_name(devname):
        """Check whether device exist."""
        if devname in BunchbyBunch.DEVICES:
            devname = devname
        elif devname in BunchbyBunch.DEVICES._fields:
            devname = BunchbyBunch.DEVICES[
                BunchbyBunch.DEVICES._fields.index(devname)
            ]
        else:
            raise NotImplementedError(devname)
        return _PVName(devname)

    def sweep_phase_shifter(self, values, wait=2, mon_type='mean'):
        """Sweep Servo Phase for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'FBE Out Phase', 'SRAM Mean'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        if self.devname.endswith('L'):
            propty = 'FBELT_SERVO_SETPT'
        elif self.devname.endswith('H'):
            propty = 'FBELT_X_PHASE_SETPT'
        else:
            propty = 'FBELT_Y_PHASE_SETPT'

        init_val = self.fbe[propty]
        for i, val in enumerate(values):
            self.fbe[propty] = val
            self.fbe.wait(propty, val)
            _time.sleep(wait)
            if mon_type.lower() in 'mean':
                mon_val = self.sram.data_mean
            else:
                mon_val = self.sram.spec_marker1_mag
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.fbe[propty] = init_val
        return _np.array(mon_values)

    def sweep_adc_delay(self, values, wait=2, mon_type='mean'):
        """Sweep ADC Delay for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'ADC Delay', 'SRAM Mean'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        init_val = self.timing.adc_delay
        for i, val in enumerate(values):
            self.timing.adc_delay = val
            self.timing.wait('TADC', val)
            _time.sleep(wait)
            if mon_type.lower() in 'mean':
                mon_val = self.sram.data_mean
            else:
                mon_val = self.sram.spec_marker1_mag
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.timing.adc_delay = init_val
        return _np.array(mon_values)

    def sweep_backend_phase(self, values, wait=2, mon_type='peak'):
        """Sweep Backend Phase for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'Backend Phase', 'Peak magnitude'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        init_val = self.fbe.be_phase
        for i, val in enumerate(values):
            self.fbe.be_phase = val
            self.fbe.wait('FBE_BE_PHASE', val)
            _time.sleep(wait)
            if mon_type.lower() in 'peak':
                mon_val = self.sram.spec_marker1_mag
            else:
                mon_val = self.sram.data_rms
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.fbe.be_phase = init_val
        return _np.array(mon_values)

    def sweep_dac_delay(self, values, wait=2, mon_type='peak'):
        """Sweep DAC Delay for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'DAC Delay', 'Peak Magnitude'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        init_val = self.timing.dac_delay
        for i, val in enumerate(values):
            self.timing.dac_delay = val
            self.timing.wait('TDAC', val)
            _time.sleep(wait)
            if mon_type.lower() in 'peak':
                mon_val = self.sram.spec_marker1_mag
            else:
                mon_val = self.sram.data_rms
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.timing.dac_delay = init_val
        return _np.array(mon_values)

    def sweep_feedback_phase(self, values, wait=2, mon_type='peak'):
        """Sweep Feedback Phase for each `value` in `values`."""
        mon_values = []
        ctrl, mon = 'Coeff. Phase', 'Peak Magnitude'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        init_val = self.coeffs.edit_phase
        for i, val in enumerate(values):
            self.coeffs.edit_phase = val
            self.coeffs.cmd_edit_apply()
            _time.sleep(wait)
            if mon_type.lower() in 'peak':
                mon_val = self.sram.spec_marker1_mag
            else:
                mon_val = self.sram.data_rms
            mon_values.append(mon_val)
            print(f'{i:03d}: {val:15.6f} {_np.mean(mon_val):15.6f}')
        self.coeffs.edit_phase = init_val
        self.coeffs.cmd_edit_apply()
        return _np.array(mon_values)

    def sweep_rf_phase(self, delta_phases, wait=2, mon_type='mean'):
        """Sweep RF Phase for each dphase in `delta_phases`."""
        mon_values = []
        ctrl, mon = 'RF Phase', 'SRAM Mean'
        print(f'Idx: {ctrl:15s} {mon:15s}')

        rfcavs = [self.rfcav_a, self.rfcav_b]
        init_phases = [rfcav.dev_llrf.phase for rfcav in rfcavs]
        for i, dphase in enumerate(delta_phases):
            for phase0, rfcav in zip(init_phases, rfcavs):
                rfcav.set_phase(phase0 + dphase)
            _time.sleep(wait)
            if mon_type.lower() in 'mean':
                mon_val = self.sram.data_mean
            else:
                mon_val = self.sram.spec_marker1_mag
            mon_values.append(mon_val)
            print(f'{i:03d}: {dphase:15.6f} {_np.mean(mon_val):15.6f}')
        for phase0, rfcav in zip(init_phases, rfcavs):
            rfcav.set_phase(phase0)
        return _np.array(mon_values)


class SystemInfo(_Device):
    """."""

    DEF_TIMEOUT = 10  # [s]

    PROPERTIES_DEFAULT = (
        'ERRSUM',
        'CLKMISS',
        'CLKMISS_COUNT',
        'PLL_UNLOCK',
        'PLL_UNLOCK_COUNT',
        'DCM_UNLOCK',
        'DCM_UNLOCK_COUNT',
        'ADC_OVR',
        'ADC_OVR_COUNT',
        'SAT',
        'SAT_COUNT',
        'FID_ERR',
        'FID_ERR_COUNT',
        'RST_COUNT',
        'CNTRST',
        'RF_FREQ',
        'FREV',
        'HARM_NUM',
        'REVISION',
        'GW_TYPE',
        'IP_ADDR',
    )

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def status(self):
        """."""
        return self['ERRSUM']

    @property
    def clock_miss(self):
        """."""
        return self['CLKMISS']

    @property
    def clock_miss_count(self):
        """."""
        return self['CLKMISS_COUNT']

    @property
    def pll_unlock(self):
        """."""
        return self['PLL_UNLOCK']

    @property
    def pll_unlock_count(self):
        """."""
        return self['PLL_UNLOCK_COUNT']

    @property
    def dcm_unlock(self):
        """."""
        return self['DCM_UNLOCK']

    @property
    def dcm_unlock_count(self):
        """."""
        return self['DCM_UNLOCK_COUNT']

    @property
    def adc_overrange(self):
        """."""
        return self['ADC_OVR']

    @property
    def adc_overrange_count(self):
        """."""
        return self['ADC_OVR_COUNT']

    @property
    def output_saturated(self):
        """."""
        return self['SAT']

    @property
    def output_saturated_count(self):
        """."""
        return self['SAT_COUNT']

    @property
    def fiducial_error(self):
        """."""
        return self['FID_ERR']

    @property
    def fiducial_error_count(self):
        """."""
        return self['FID_ERR_COUNT']

    @property
    def time_since_last_reset(self):
        """."""
        return self['RST_COUNT']

    def cmd_reset_counts(self, timeout=DEF_TIMEOUT):
        """."""
        self['CNTRST'] = 1
        if not self.wait('CNTRST', 1, timeout / 2):
            return False
        _time.sleep(0.2)
        self['CNTRST'] = 0
        return self.wait('CNTRST', 0, timeout / 2)

    @property
    def rf_freq_nom(self):
        """."""
        return self['RF_FREQ'] * 1e6

    @property
    def revolution_freq_nom(self):
        """."""
        return self['FREV'] * 1e3

    @property
    def harmonic_number(self):
        """."""
        return self['HARM_NUM']

    @property
    def gateway_revision(self):
        """."""
        return self['REVISION']

    @property
    def gateway_type(self):
        """."""
        return self['GW_TYPE']

    @property
    def ip_address(self):
        """."""
        return self['IP_ADDR']


class Timing(_Device):
    """."""

    DEF_TIMEOUT = 10  # [s]

    PROPERTIES_DEFAULT = (
        'TADC',
        'TDAC',
        'DELAY',
        'OFF_FIDS',
        'FID_DELAY',
        'CLKRST',
        'FREQ_CNT_CH0',
        'FREQ_CNT_CH1',
        'FREQ_CNT_CH2',
        'FREQ_CNT_CH3',
        'FREQ_CNT_CH4',
        'ECLDEL0',
        'ECLDEL1',
        'ECLDEL2',
        'ECLDEL3',
        'ECLDEL0_SUBWR',
        'ECLDEL1_SUBWR',
        'ECLDEL2_SUBWR',
        'ECLDEL3_SUBWR',
    )

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def adc_delay(self):
        """."""
        return self['TADC']

    @adc_delay.setter
    def adc_delay(self, value):
        self['TADC'] = value

    @property
    def dac_delay(self):
        """."""
        return self['TDAC']

    @dac_delay.setter
    def dac_delay(self, value):
        self['TDAC'] = value

    @property
    def output_delay(self):
        """."""
        return self['DELAY']

    @output_delay.setter
    def output_delay(self, value):
        self['DELAY'] = value

    @property
    def fiducial_offset(self):
        """."""
        return self['OFF_FIDS']

    @fiducial_offset.setter
    def fiducial_offset(self, value):
        self['OFF_FIDS'] = int(value)

    @property
    def fiducial_delay(self):
        """."""
        return self['FID_DELAY']

    @fiducial_delay.setter
    def fiducial_delay(self, value):
        self['FID_DELAY'] = value

    def cmd_reset_clock(self, timeout=DEF_TIMEOUT):
        """."""
        self['CLKRST'] = 1
        if not self.wait('CLKRST', 1, timeout / 2):
            return False
        _time.sleep(0.2)
        self['CLKRST'] = 0
        return self.wait('CLKRST', 0, timeout / 2)

    @property
    def adc_clock(self):
        """."""
        return self['ECLDEL0']

    @adc_clock.setter
    def adc_clock(self, value):
        self['ECLDEL0'] = int(value)

    @property
    def fiducial_clock(self):
        """."""
        return self['ECLDEL1']

    @fiducial_clock.setter
    def fiducial_clock(self, value):
        self['ECLDEL1'] = int(value)

    @property
    def fiducial(self):
        """."""
        return self['ECLDEL2']

    @fiducial.setter
    def fiducial(self, value):
        self['ECLDEL2'] = int(value)

    @property
    def dac_clock(self):
        """."""
        return self['ECLDEL3']

    @dac_clock.setter
    def dac_clock(self, value):
        self['ECLDEL3'] = int(value)

    @property
    def adc_clock_ok(self):
        """."""
        return self['ECLDEL0_SUBWR']

    @property
    def fiducial_clock_ok(self):
        """."""
        return self['ECLDEL1_SUBWR']

    @property
    def fiducial_ok(self):
        """."""
        return self['ECLDEL2_SUBWR']

    @property
    def dac_clock_ok(self):
        """."""
        return self['ECLDEL3_SUBWR']

    @property
    def freq_input_clock(self):
        """."""
        return self['FREQ_CNT_CH0']

    @property
    def freq_rf_over_2(self):
        """."""
        return self['FREQ_CNT_CH1']

    @property
    def freq_rf_over_3(self):
        """."""
        return self['FREQ_CNT_CH2']

    @property
    def freq_dac_clock(self):
        """."""
        return self['FREQ_CNT_CH3']

    @property
    def freq_rf_over_4(self):
        """."""
        return self['FREQ_CNT_CH4']


class Coefficients(_Device):
    """."""

    PROPERTIES_DEFAULT = (
        'CSET0',
        'CSET1',
        'CSET2',
        'CSET3',
        'COEFF',
        'DESC_CSET0',
        'DESC_CSET1',
        'DESC_CSET2',
        'DESC_CSET3',
        'DESC_COEFF',
        'CVERIFY.C',
        'CVERIFY.D',
        'CVERIFY.G',
        'CVERIFY.H',
        'LDSET',
        'BO_CVERIFY',
        'BO_CPCOEFF',
        'FLT_GAIN',
        'FLT_FREQ',
        'FLT_PHASE',
        'FLT_TAPS',
        'FTF_MAG',
        'FTF_PHASE',
        'FTF_FREQ',
        'FTF_TUNE',
        'FTF_GTUNE',
        'FTF_PTUNE',
    )

    DEF_TIMEOUT = 10  # [s]
    OFF, ON = 0, 1
    FPGA_BITS = 2**15

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def set0(self):
        """."""
        return self['CSET0'] / self.FPGA_BITS

    @set0.setter
    def set0(self, value):
        self['CSET0'] = _np.array(value) * self.FPGA_BITS

    @property
    def set1(self):
        """."""
        return self['CSET1'] / self.FPGA_BITS

    @set1.setter
    def set1(self, value):
        self['CSET1'] = _np.array(value) * self.FPGA_BITS

    @property
    def set2(self):
        """."""
        return self['CSET2'] / self.FPGA_BITS

    @set2.setter
    def set2(self, value):
        self['CSET2'] = _np.array(value) * self.FPGA_BITS

    @property
    def set3(self):
        """."""
        return self['CSET3'] / self.FPGA_BITS

    @set3.setter
    def set3(self, value):
        self['CSET3'] = _np.array(value) * self.FPGA_BITS

    @property
    def set0_desc(self):
        """."""
        return self['DESC_CSET0']

    @property
    def set1_desc(self):
        """."""
        return self['DESC_CSET1']

    @property
    def set2_desc(self):
        """."""
        return self['DESC_CSET2']

    @property
    def set3_desc(self):
        """."""
        return self['DESC_CSET3']

    @property
    def set0_status(self):
        """."""
        return self['CVERIFY.C']

    @property
    def set1_status(self):
        """."""
        return self['CVERIFY.D']

    @property
    def set2_status(self):
        """."""
        return self['CVERIFY.G']

    @property
    def set3_status(self):
        """."""
        return self['CVERIFY.H']

    @property
    def edit_choose_set(self):
        """."""
        return self['LDSET']

    @edit_choose_set.setter
    def edit_choose_set(self, value):
        """."""
        self['LDSET'] = value

    @property
    def edit_set(self):
        """."""
        return self['COEFF'] / self.FPGA_BITS

    @edit_set.setter
    def edit_set(self, value):
        """."""
        self['COEFF'] = _np.array(value) * self.FPGA_BITS

    @property
    def edit_set_desc(self):
        """."""
        return self['DESC_COEFF']

    @property
    def edit_set_status(self):
        """."""
        return self['BO_CVERIFY']

    @property
    def edit_ftf_magnitude(self):
        """."""
        return self['FTF_MAG']

    @property
    def edit_ftf_phase(self):
        """."""
        return self['FTF_PHASE']

    @property
    def edit_ftf_freq(self):
        """."""
        return self['FTF_FREQ']

    @property
    def edit_marker_freq(self):
        """."""
        return self['FTF_TUNE']

    @property
    def edit_marker_mag(self):
        """."""
        return self['FTF_GTUNE'][0]

    @property
    def edit_marker_phase(self):
        """."""
        return self['FTF_PTUNE'][0]

    @property
    def edit_gain(self):
        """."""
        return self['FLT_GAIN']

    @edit_gain.setter
    def edit_gain(self, value):
        """."""
        self['FLT_GAIN'] = value

    @property
    def edit_freq(self):
        """."""
        return self['FLT_FREQ']

    @edit_freq.setter
    def edit_freq(self, value):
        """."""
        self['FLT_FREQ'] = value

    @property
    def edit_phase(self):
        """."""
        return self['FLT_PHASE']

    @edit_phase.setter
    def edit_phase(self, value):
        """."""
        self['FLT_PHASE'] = value

    @property
    def edit_nrpoints(self):
        """."""
        return self['FLT_TAPS']

    @edit_nrpoints.setter
    def edit_nrpoints(self, value):
        """."""
        self['FLT_TAPS'] = value

    def cmd_edit_apply(self, timeout=DEF_TIMEOUT):
        """."""
        self['BO_CPCOEFF'] = 1
        return self.wait('BO_CPCOEFF', 1, timeout)

    def cmd_edit_verify(self, timeout=DEF_TIMEOUT):
        """."""
        self['BO_CVERIFY'] = 1
        return self.wait('BO_CVERIFY', 1, timeout)


class Acquisition(_Device):
    """."""

    ACQTYPE = _get_namedtuple('Devices', ('SRAM', 'BRAM'))

    PROPERTIES_DEFAULT = (
        'GDTIME',
        'HOLDTIME',
        'POSTTIME',
        'ACQTIME',
        'REC_DS',
        'POSTSEL',
        'ACQ_EN',
        'ACQ_SINGLE',
        'SP_AVG',
        'HWTEN',
        'TRIG_IN_SEL',
        'ARM',
        'ARM_MON',
        'BR_ARM',
        'DUMP',
        'RAW_SAMPLES',
        'RAW',
        'ACQ_TURNS',
        'POST_TURNS',
        'MEAN',
        'RMS',
        'XSC',
        'SPEC',
        'MAXRMS',
        'TSC',
        'FREQ',
        'ACQ_MASK',
        'ACQ_PATTERN',
        'SP_LOW1',
        'SP_HIGH1',
        'SP_SEARCH1',
        'SP_LOW2',
        'SP_HIGH2',
        'SP_SEARCH2',
        'PEAKFREQ1',
        'PEAK1',
        'PEAKTUNE1',
        'PEAKFREQ2',
        'PEAK2',
        'PEAKTUNE2',
        'MD_ENABLE',
        'MD_SMODE',
        'MD_FTUNE',
        'MD_FSPAN',
        'MD_MSEL',
        'MD_AVG',
        'MD_SP_LOW',
        'MD_SP_HIGH',
        'MD_SP_SEARCH',
        'MD_MAXMODE',
        'MD_MAXVAL',
        'MD_PEAK',
        'MD_PEAKFREQ',
        'MD_PEAKTUNE',
        'MD_MODES',
        'MD_SPEC',
    )

    DEF_TIMEOUT = 10  # [s]

    def __init__(self, devname, acqtype='BRAM', props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        acqtype = Acquisition.process_acquisition_type(acqtype)
        super().__init__(devname + ':' + acqtype + '_', props2init=props2init)

        pvo = self.pv_object('RAW')
        self._update_data_evt = _Event()
        pvo.auto_monitor = True
        pvo.add_callback(self._update_evt)

    # ########### Acquisition Config Properties ###########
    @property
    def acq_type(self):
        """."""
        return self['POSTSEL']

    @acq_type.setter
    def acq_type(self, value):
        self['POSTSEL'] = value

    @property
    def acq_enbl(self):
        """."""
        return self['ACQ_EN']

    @acq_enbl.setter
    def acq_enbl(self, value):
        self['ACQ_EN'] = value

    @property
    def acq_mode(self):
        """."""
        return self['ACQ_SINGLE']

    @acq_mode.setter
    def acq_mode(self, value):
        self['ACQ_SINGLE'] = value

    @property
    def growthtime(self):
        """."""
        return self['GDTIME']

    @growthtime.setter
    def growthtime(self, value):
        self['GDTIME'] = value

    @property
    def acqtime(self):
        """."""
        return self['ACQTIME']

    @acqtime.setter
    def acqtime(self, value):
        self['ACQTIME'] = value

    @property
    def holdtime(self):
        """."""
        return self['HOLDTIME']

    @holdtime.setter
    def holdtime(self, value):
        self['HOLDTIME'] = value

    @property
    def posttime(self):
        """."""
        return self['POSTTIME']

    @posttime.setter
    def posttime(self, value):
        self['POSTTIME'] = value

    @property
    def downsample(self):
        """."""
        return self['REC_DS']

    @downsample.setter
    def downsample(self, value):
        self['REC_DS'] = value

    @property
    def nr_averages(self):
        """."""
        return self['SP_AVG']

    @nr_averages.setter
    def nr_averages(self, value):
        self['SP_AVG'] = value

    # ########### Trigger Properties ###########
    @property
    def trigger_type(self):
        """."""
        return self['HWTEN']

    @trigger_type.setter
    def trigger_type(self, value):
        self['HWTEN'] = value

    @property
    def trigger_sel(self):
        """."""
        return self['TRIG_IN_SEL']

    @trigger_sel.setter
    def trigger_sel(self, value):
        self['TRIG_IN_SEL'] = value

    @property
    def trigger_armed(self):
        """."""
        return self['ARM_MON']

    @trigger_armed.setter
    def trigger_armed(self, value):
        self['ARM'] = value

    @property
    def trigger_rearm(self):
        """."""
        return self['BR_ARM']

    @trigger_rearm.setter
    def trigger_rearm(self, value):
        self['BR_ARM'] = value

    # ########### Data Properties ###########
    @property
    def data_size(self):
        """."""
        return self['RAW_SAMPLES']

    @data_size.setter
    def data_size(self, value):
        self['RAW_SAMPLES'] = value

    @property
    def data_raw(self):
        """."""
        data = self['RAW']
        size = self.data_size
        if size is not None:
            data = data[:size]
        return data

    @property
    def data_bunch_index(self):
        """."""
        return self['XSC']

    @property
    def data_mean(self):
        """."""
        return self['MEAN']

    @property
    def data_rms(self):
        """."""
        return self['RMS']

    @property
    def data_nrturns(self):
        """."""
        return self['ACQ_TURNS']

    @property
    def data_nrturns_posttriger(self):
        """."""
        return self['POST_TURNS']

    def cmd_data_acquire(self, timeout=DEF_TIMEOUT):
        """."""
        self.acq_enbl = 1
        return self.wait('ACQ_EN', 1, timeout=timeout)

    def cmd_data_dump(self, timeout=DEF_TIMEOUT, pv_update=False):
        """."""
        if pv_update:
            self._update_data_evt.clear()
        self['DUMP'] = 1
        return self.wait_data_dump(timeout, pv_update=pv_update)

    def wait_data_dump(self, timeout=None, pv_update=False):
        """."""
        timeout = timeout or Acquisition.DEF_TIMEOUT
        if not self.wait('DUMP', False, timeout=timeout):
            print('WARN: Timed out waiting data dump.')
            return False
        if pv_update and not self._update_data_evt.wait(timeout=timeout):
            print('WARN: Timed out waiting PV data update.')
            return False
        self._update_data_evt.clear()
        return True

    # ########### Spectrometer Properties ###########
    @property
    def spec_mag(self):
        """."""
        return self['SPEC']

    @property
    def spec_freq(self):
        """."""
        return self['FREQ']

    @property
    def spec_data(self):
        """."""
        return self['MAXRMS']

    @property
    def spec_time(self):
        """."""
        return self['TSC']

    @property
    def spec_mask(self):
        """."""
        return self['ACQ_MASK']

    @spec_mask.setter
    def spec_mask(self, value):
        self['ACQ_MASK'] = _np.array(value)

    @property
    def spec_mask_pattern(self):
        """."""
        return self['ACQ_PATTERN']

    @spec_mask_pattern.setter
    def spec_mask_pattern(self, value):
        self['ACQ_PATTERN'] = str(value)

    @property
    def spec_marker1_freq_min(self):
        """."""
        return self['SP_LOW1']

    @spec_marker1_freq_min.setter
    def spec_marker1_freq_min(self, value):
        self['SP_LOW1'] = value

    @property
    def spec_marker1_freq_max(self):
        """."""
        return self['SP_HIGH1']

    @spec_marker1_freq_max.setter
    def spec_marker1_freq_max(self, value):
        self['SP_HIGH1'] = value

    @property
    def spec_marker1_search_mode(self):
        """."""
        return self['SP_SEARCH1']

    @spec_marker1_search_mode.setter
    def spec_marker1_search_mode(self, value):
        self['SP_SEARCH1'] = value

    @property
    def spec_marker1_freq(self):
        """."""
        return self['PEAKFREQ1']

    @property
    def spec_marker1_tune(self):
        """."""
        return self['PEAKTUNE1']

    @property
    def spec_marker1_mag(self):
        """."""
        return self['PEAK1']

    @property
    def spec_marker2_freq_min(self):
        """."""
        return self['SP_LOW2']

    @spec_marker2_freq_min.setter
    def spec_marker2_freq_min(self, value):
        self['SP_LOW2'] = value

    @property
    def spec_marker2_freq_max(self):
        """."""
        return self['SP_HIGH2']

    @spec_marker2_freq_max.setter
    def spec_marker2_freq_max(self, value):
        self['SP_HIGH2'] = value

    @property
    def spec_marker2_search_mode(self):
        """."""
        return self['SP_SEARCH2']

    @spec_marker2_search_mode.setter
    def spec_marker2_search_mode(self, value):
        self['SP_SEARCH2'] = value

    @property
    def spec_marker2_freq(self):
        """."""
        return self['PEAKFREQ2']

    @property
    def spec_marker2_tune(self):
        """."""
        return self['PEAKTUNE2']

    @property
    def spec_marker2_mag(self):
        """."""
        return self['PEAK2']

    @staticmethod
    def process_acquisition_type(acqtype):
        """."""
        if acqtype in Acquisition.ACQTYPE:
            acqtype = Acquisition.ACQTYPE._fields[acqtype]
        elif acqtype in Acquisition.ACQTYPE._fields:
            acqtype = acqtype
        else:
            raise NotImplementedError(acqtype)
        return acqtype

    # ############# Modal Analysis Properties #############
    @property
    def modal_acq_enbl(self):
        """."""
        return self['MD_ENABLE']

    @modal_acq_enbl.setter
    def modal_acq_enbl(self, value):
        self['MD_ENABLE'] = value

    @property
    def modal_acq_mode(self):
        """."""
        return self['MD_SMODE']

    @modal_acq_mode.setter
    def modal_acq_mode(self, value):
        self['MD_SMODE'] = value

    @property
    def modal_mode_selected(self):
        """."""
        return self['MD_MSEL']

    @modal_mode_selected.setter
    def modal_mode_selected(self, value):
        self['MD_MSEL'] = value

    @property
    def modal_nr_averages(self):
        """."""
        return self['MD_AVG']

    @modal_nr_averages.setter
    def modal_nr_averages(self, value):
        self['MD_AVG'] = value

    @property
    def modal_sideband_freq(self):
        """."""
        return self['MD_FTUNE']

    @modal_sideband_freq.setter
    def modal_sideband_freq(self, value):
        self['MD_FTUNE'] = value

    @property
    def modal_sideband_span(self):
        """."""
        return self['MD_FSPAN']

    @modal_sideband_span.setter
    def modal_sideband_span(self, value):
        self['MD_FSPAN'] = value

    @property
    def modal_marker_freq_min(self):
        """."""
        return self['MD_SP_LOW']

    @modal_marker_freq_min.setter
    def modal_marker_freq_min(self, value):
        self['MD_SP_LOW'] = value

    @property
    def modal_marker_freq_max(self):
        """."""
        return self['MD_SP_HIGH']

    @modal_marker_freq_max.setter
    def modal_marker_freq_max(self, value):
        self['MD_SP_HIGH'] = value

    @property
    def modal_marker_search_mode(self):
        """."""
        return self['MD_SP_SEARCH']

    @modal_marker_search_mode.setter
    def modal_marker_search_mode(self, value):
        self['MD_SP_SEARCH'] = value

    @property
    def modal_maximum_mode(self):
        """."""
        return self['MD_MAXMODE']

    @property
    def modal_maximum_value(self):
        """."""
        return self['MD_MAXVAL']

    @property
    def modal_marker_freq(self):
        """."""
        return self['MD_PEAKFREQ']

    @property
    def modal_marker_tune(self):
        """."""
        return self['MD_PEAKTUNE']

    @property
    def modal_marker_mag(self):
        """."""
        return self['MD_PEAK']

    @property
    def modal_modes_amp(self):
        """."""
        return self['MD_MODES']

    @property
    def modal_maximum_mode_spec(self):
        """."""
        return self['MD_SPEC']

    def _update_evt(self, **kwargs):
        _ = kwargs
        self._update_data_evt.set()


class SingleBunch(_Device):
    """."""

    PROPERTIES_DEFAULT = (
        'ACQTIME',
        'ACQ_SAMPLES',
        'ACQ_EN',
        'ACQ_SINGLE',
        'BUNCH_ID',
        'RAW_BUNCH_ID',
        'EXTEN',
        'TRIG_IN_SEL',
        'ARM',
        'ARM_MON',
        'BR_ARM',
        'RAW_SAMPLES',
        'TSC',
        'RAW',
        'FREQ',
        'MAG',
        'PHASE',
        'TF_ENABLE',
        'NFFT',
        'NOVERLAP',
        'DEL_CAL',
        'SP_AVG',
        'MEANVAL',
        'RMSVAL',
        'AMP_PP',
        'SP_LOW1',
        'SP_HIGH1',
        'PEAKFREQ1',
        'PEAKTUNE1',
        'PEAK1',
        'SP_SEARCH1',
        'PHASE1',
    )

    DEF_TIMEOUT = 10  # [s]

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname + ':' + 'SB_', props2init=props2init)

    @property
    def acqtime(self):
        """."""
        return self['ACQTIME']

    @acqtime.setter
    def acqtime(self, value):
        self['ACQTIME'] = value

    @property
    def acq_enbl(self):
        """."""
        return self['ACQ_EN']

    @acq_enbl.setter
    def acq_enbl(self, value):
        self['ACQ_EN'] = value

    @property
    def acq_mode(self):
        """."""
        return self['ACQ_SINGLE']

    @acq_mode.setter
    def acq_mode(self, value):
        self['ACQ_SINGLE'] = value

    @property
    def bunch_id(self):
        """."""
        return self['RAW_BUNCH_ID']

    @bunch_id.setter
    def bunch_id(self, value):
        self['BUNCH_ID'] = value

    # ########### Trigger Properties ###########
    @property
    def trigger_type(self):
        """."""
        return self['EXTEN']

    @trigger_type.setter
    def trigger_type(self, value):
        self['EXTEN'] = value

    @property
    def trigger_sel(self):
        """."""
        return self['TRIG_IN_SEL']

    @trigger_sel.setter
    def trigger_sel(self, value):
        self['TRIG_IN_SEL'] = value

    @property
    def trigger_armed(self):
        """."""
        return self['ARM_MON']

    @trigger_armed.setter
    def trigger_armed(self, value):
        self['ARM'] = value

    @property
    def trigger_rearm(self):
        """."""
        return self['BR_ARM']

    @trigger_rearm.setter
    def trigger_rearm(self, value):
        self['BR_ARM'] = value

    # ########### Data Properties ###########
    @property
    def data_size(self):
        """."""
        return self['RAW_SAMPLES']

    @data_size.setter
    def data_size(self, value):
        self['RAW_SAMPLES'] = value

    @property
    def data_raw(self):
        """."""
        data = self['RAW']
        size = self.data_size
        if size is not None:
            data = data[:size]
        return data

    @property
    def data_time(self):
        """."""
        return self['TSC']

    ##
    @property
    def fft_size(self):
        """."""
        return self['NFFT']

    @fft_size.setter
    def fft_size(self, value):
        self['NFFT'] = value

    @property
    def fft_overlap(self):
        """."""
        return self['NOVERLAP']

    @fft_overlap.setter
    def fft_overlap(self, value):
        self['NOVERLAP'] = value

    @property
    def delay_calibration(self):
        """."""
        return self['DEL_CAL']

    @delay_calibration.setter
    def delay_calibration(self, value):
        self['DEL_CAL'] = value

    @property
    def nr_averages(self):
        """."""
        return self['SP_AVG']

    @nr_averages.setter
    def nr_averages(self, value):
        self['SP_AVG'] = value

    @property
    def transfer_function_enable(self):
        """."""
        return self['TF_ENABLE']

    @transfer_function_enable.setter
    def transfer_function_enable(self, value):
        self['TF_ENABLE'] = value

    @property
    def amplitude_mean(self):
        """."""
        return self['MEANVAL']

    @amplitude_mean.setter
    def amplitude_mean(self, value):
        self['MEANVAL'] = value

    @property
    def amplitude_rms(self):
        """."""
        return self['RMSVAL']

    @amplitude_rms.setter
    def amplitude_rms(self, value):
        self['RMSVAL'] = value

    @property
    def amplitude_pp(self):
        """."""
        return self['AMP_PP']

    @amplitude_pp.setter
    def amplitude_pp(self, value):
        self['AMP_PP'] = value

    @property
    def spec_mag(self):
        """."""
        return self['MAG']

    @property
    def spec_phase(self):
        """."""
        return self['PHASE']

    @property
    def spec_freq(self):
        """."""
        return self['FREQ']

    @property
    def spec_marker1_freq_min(self):
        """."""
        return self['SP_LOW1']

    @spec_marker1_freq_min.setter
    def spec_marker1_freq_min(self, value):
        self['SP_LOW1'] = value

    @property
    def spec_marker1_freq_max(self):
        """."""
        return self['SP_HIGH1']

    @spec_marker1_freq_max.setter
    def spec_marker1_freq_max(self, value):
        self['SP_HIGH1'] = value

    @property
    def spec_marker1_search_mode(self):
        """."""
        return self['SP_SEARCH1']

    @spec_marker1_search_mode.setter
    def spec_marker1_search_mode(self, value):
        self['SP_SEARCH1'] = value

    @property
    def spec_marker1_freq(self):
        """."""
        return self['PEAKFREQ1']

    @property
    def spec_marker1_tune(self):
        """."""
        return self['PEAKTUNE1']

    @property
    def spec_marker1_mag(self):
        """."""
        return self['PEAK1']

    @property
    def spec_marker1_phase(self):
        """."""
        return self['PHASE1']

    def cmd_enable_transfer_function(self, timeout=DEF_TIMEOUT):
        """Enable transfer function."""
        self.transfer_function_enable = 1
        return self.wait('TF_ENABLE', value=1, timeout=timeout)


class PhaseTracking(_Device):
    """."""

    PROPERTIES_DEFAULT = (
        'PHTRK_GAIN',
        'PHTRK_SETPT',
        'PHTRK_RANGE',
        'PHTRK_DECIM',
        'PHTRK_RATE',
        'PHTRK_BANDWIDTH',
        'PHTRK_LOOPCTRL',
        'PHTRK_MAG',
        'PHTRK_TFGAIN',
        'PHTRK_SHIFT',
        'PHTRK_PHASE',
        'PHTRK_ERROR',
        'PHTRK_FREQ0',
        'PHTRK_TUNE',
        'DRIVE2_TRACK',
        'PHTRK_MOD',
    )

    DEF_TIMEOUT = 10  # [s]

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def gain(self):
        """."""
        return self['PHTRK_GAIN']

    @gain.setter
    def gain(self, value):
        self['PHTRK_GAIN'] = value

    @property
    def setpoint(self):
        """Phase setpoint in degrees."""
        return self['PHTRK_SETPT']

    @setpoint.setter
    def setpoint(self, value):
        self['PHTRK_SETPT'] = value

    @property
    def range(self):
        """Phase tracking frequency range [kHz]."""
        return self['PHTRK_RANGE']

    @range.setter
    def range(self, value):
        self['PHTRK_RANGE'] = value

    @property
    def decimation(self):
        """."""
        return self['PHTRK_DECIM']

    @decimation.setter
    def decimation(self, value):
        self['PHTRK_DECIM'] = value

    @property
    def loop_state(self):
        """Loop State."""
        return self['PHTRK_LOOPCTRL']

    @loop_state.setter
    def loop_state(self, value):
        self['PHTRK_LOOPCTRL'] = value

    @property
    def use_drive2(self):
        """Loop State."""
        return self['DRIVE2_TRACK']

    @use_drive2.setter
    def use_drive2(self, value):
        self['DRIVE2_TRACK'] = value

    @property
    def time_modulation(self):
        """Loop State."""
        return self['PHTRK_MOD']

    @time_modulation.setter
    def time_modulation(self, value):
        self['PHTRK_MOD'] = value

    @property
    def rate(self):
        """Phase tracking rate control [kHz]."""
        return self['PHTRK_RATE']

    @property
    def bandwidth(self):
        """Phase tracking bandwidth control [Hz]."""
        return self['PHTRK_BANDWIDTH']

    @property
    def magnitude(self):
        """Phase tracking data magnitude [counts]."""
        return self['PHTRK_MAG']

    @property
    def transfer_gain(self):
        """Phase tracking data transfer gain [dB]."""
        return self['PHTRK_TFGAIN']

    @property
    def normalizing_shift(self):
        """Phase tracking data normalizing shift."""
        return self['PHTRK_SHIFT']

    @property
    def phase(self):
        """Phase tracking data monitor [deg]."""
        return self['PHTRK_PHASE']

    @property
    def error(self):
        """Phase tracking data error [deg]."""
        return self['PHTRK_ERROR']

    @property
    def frequency(self):
        """Phase tracking data frequency [kHz]."""
        return self['PHTRK_FREQ0']

    @property
    def tune(self):
        """Phase tracking data tune."""
        return self['PHTRK_TUNE']


class FrontBackEnd(_Device):
    """."""

    PROPERTIES_DEFAULT = (
        'FBE_Z_ATT',
        'FBE_Z_PHASE',
        'FBELT_SERVO_SETPT',
        'FBE_BE_ATT',
        'FBE_BE_PHASE',
        'FBE_X_ATT',
        'FBE_X_PHASE',
        'FBELT_X_PHASE_SETPT',
        'FBE_Y_ATT',
        'FBE_Y_PHASE',
        'FBELT_Y_PHASE_SETPT',
    )

    FPGA_BITS = 2**15

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def z_att(self):
        """."""
        return self['FBE_Z_ATT']

    @z_att.setter
    def z_att(self, value):
        """."""
        self['FBE_Z_ATT'] = value

    @property
    def be_att(self):
        """."""
        return self['FBE_BE_ATT']

    @be_att.setter
    def be_att(self, value):
        """."""
        self['FBE_BE_ATT'] = value

    @property
    def x_att(self):
        """."""
        return self['FBE_X_ATT']

    @x_att.setter
    def x_att(self, value):
        """."""
        self['FBE_X_ATT'] = value

    @property
    def y_att(self):
        """."""
        return self['FBE_Y_ATT']

    @y_att.setter
    def y_att(self, value):
        """."""
        self['FBE_Y_ATT'] = value

    @property
    def z_phase(self):
        """."""
        return self['FBE_Z_PHASE']

    @z_phase.setter
    def z_phase(self, value):
        """."""
        self['FBELT_SERVO_SETPT'] = value

    @property
    def be_phase(self):
        """."""
        return self['FBE_BE_PHASE']

    @be_phase.setter
    def be_phase(self, value):
        """."""
        self['FBE_BE_PHASE'] = value

    @property
    def x_phase(self):
        """."""
        return self['FBE_X_PHASE']

    @x_phase.setter
    def x_phase(self, value):
        """."""
        self['FBELT_X_PHASE_SETPT'] = value

    @property
    def y_phase(self):
        """."""
        return self['FBE_Y_PHASE']

    @y_phase.setter
    def y_phase(self, value):
        """."""
        self['FBELT_Y_PHASE_SETPT'] = value


class Feedback(_Device):
    """."""

    PROPERTIES_DEFAULT = (
        'PROC_DS',
        'FBCTRL',
        'SHIFTGAIN',
        'SETSEL',
        'SAT_THRESHOLD',
        'FB_MASK',
        'FB_PATTERN',
        'CF_MASK',
        'CF_PATTERN',
        'CF_PATTERN_SUB.VALB',
        'GDEN',
    )

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname, props2init=props2init)

    @property
    def downsample(self):
        """."""
        return self['PROC_DS']

    @downsample.setter
    def downsample(self, value):
        self['PROC_DS'] = int(value)

    @property
    def loop_state(self):
        """."""
        return self['FBCTRL']

    @loop_state.setter
    def loop_state(self, value):
        self['FBCTRL'] = int(value)

    @property
    def shift_gain(self):
        """."""
        return self['SHIFTGAIN']

    @shift_gain.setter
    def shift_gain(self, value):
        self['SHIFTGAIN'] = int(value)

    @property
    def coeff_set(self):
        """."""
        return self['SETSEL']

    @coeff_set.setter
    def coeff_set(self, value):
        self['SETSEL'] = int(value)

    @property
    def saturation_threshold(self):
        """."""
        return self['SAT_THRESHOLD']

    @saturation_threshold.setter
    def saturation_threshold(self, value):
        self['SAT_THRESHOLD'] = value

    @property
    def mask(self):
        """."""
        return self['FB_MASK']

    @mask.setter
    def mask(self, value):
        self['FB_MASK'] = _np.array(value)

    @property
    def mask_pattern(self):
        """."""
        return self['FB_PATTERN']

    @mask_pattern.setter
    def mask_pattern(self, value):
        self['FB_PATTERN'] = str(value)

    @property
    def alternate_mask(self):
        """."""
        return self['CF_MASK']

    @alternate_mask.setter
    def alternate_mask(self, value):
        self['CF_MASK'] = _np.array(value)

    @property
    def alternate_mask_pattern(self):
        """."""
        return self['CF_PATTERN']

    @alternate_mask_pattern.setter
    def alternate_mask_pattern(self, value):
        self['CF_PATTERN'] = str(value)

    @property
    def alternate_nr_bunches(self):
        """."""
        return self['CF_PATTERN_SUB.VALB']

    @property
    def alternate_inuse(self):
        """."""
        return bool(self.alternate_nr_bunches)

    @property
    def grow_damp_state(self):
        """."""
        return self['GDEN']

    @grow_damp_state.setter
    def grow_damp_state(self, value):
        self['GDEN'] = int(value)


class Drive(_Device):
    """."""

    PROPERTIES_DEFAULT = (
        'MOD',
        'AMPL',
        'WAVEFORM',
        'FREQ',
        'FREQ_ACT',
        'SPAN',
        'SPAN_ACT',
        'PERIOD',
        'PERIOD_ACT',
        'MASK',
        'PATTERN',
        'BITS',
    )

    def __init__(self, devname, drive_num=None, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        propty = 'DRIVE'
        if drive_num is not None:
            propty += str(drive_num)
        propty += '_'
        super().__init__(devname + ':' + propty, props2init=props2init)

    @property
    def number_of_bits(self):
        """."""
        return self['BITS']

    @property
    def state(self):
        """."""
        return self['MOD']

    @state.setter
    def state(self, value):
        self['MOD'] = int(value)

    @property
    def waveform(self):
        """."""
        return self['WAVEFORM']

    @waveform.setter
    def waveform(self, value):
        self['WAVEFORM'] = int(value)

    @property
    def amplitude(self):
        """."""
        return self['AMPL']

    @amplitude.setter
    def amplitude(self, value):
        self['AMPL'] = value

    @property
    def frequency(self):
        """Frequency [kHz]."""
        return self['FREQ_ACT'] / 1e3

    @frequency.setter
    def frequency(self, value):
        self['FREQ'] = value

    @property
    def span(self):
        """."""
        return self['SPAN_ACT']

    @span.setter
    def span(self, value):
        self['SPAN'] = value

    @property
    def period(self):
        """."""
        return self['PERIOD_ACT']

    @period.setter
    def period(self, value):
        self['PERIOD'] = value

    @property
    def mask(self):
        """."""
        return self['MASK']

    @mask.setter
    def mask(self, value):
        self['MASK'] = _np.array(value)

    @property
    def mask_pattern(self):
        """."""
        return self['PATTERN']

    @mask_pattern.setter
    def mask_pattern(self, value):
        self['PATTERN'] = str(value)


class BunchClean(_Device):
    """."""

    PROPERTIES_DEFAULT = (
        'ENABLE',
        'AMPL',
        'TUNE',
        'PATTERN',
        'PERIOD',
        'SPAN',
    )

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname + ':CLEAN_', props2init=props2init)

    @property
    def state(self):
        """."""
        return self['ENABLE']

    @state.setter
    def state(self, value):
        self['ENABLE'] = int(value)

    @property
    def amplitude(self):
        """."""
        return self['AMPL']

    @amplitude.setter
    def amplitude(self, value):
        self['AMPL'] = value

    @property
    def frequency(self):
        """."""
        return self['TUNE']

    @frequency.setter
    def frequency(self, value):
        self['TUNE'] = value

    @property
    def span(self):
        """."""
        return self['SPAN']

    @span.setter
    def span(self, value):
        self['SPAN'] = value

    @property
    def period(self):
        """."""
        return self['PERIOD']

    @period.setter
    def period(self, value):
        self['PERIOD'] = value

    @property
    def mask_pattern(self):
        """."""
        return self['PATTERN']

    @mask_pattern.setter
    def mask_pattern(self, value):
        self['PATTERN'] = str(value)


class PwrAmpL(_Device):
    """."""

    PROPERTIES_DEFAULT = ('FAULT', 'TEMP', 'FWDLOSS', 'REVLOSS', 'FWD', 'REV')

    DEF_TIMEOUT = 10  # [s]

    def __init__(self, devname, num=0, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        super().__init__(devname + f':MCLRAW_{num:d}_', props2init=props2init)

    @property
    def status(self):
        """."""
        return self['FAULT']

    @property
    def forward_loss(self):
        """."""
        return self['FWDLOSS']

    @forward_loss.setter
    def forward_loss(self, value):
        self['FWDLOSS'] = value

    @property
    def reverse_loss(self):
        """."""
        return self['REVLOSS']

    @reverse_loss.setter
    def reverse_loss(self, value):
        self['REVLOSS'] = value

    @property
    def forward(self):
        """."""
        return self['FWD']

    @property
    def reverse(self):
        """."""
        return self['REV']

    @property
    def temperature(self):
        """."""
        return self['TEMP']


class PwrAmpT(_Device):
    """."""

    DEF_TIMEOUT = 10  # [s]

    PROPERTIES_DEFAULT = (
        'Rst-Cmd',
        'Enbl-Sts',
        'Enbl-Sel',
        'GainAuto-Sts',
        'GainAuto-Sel',
        'Gain-SP',
        'Gain-RB',
        'GainStep-SP',
        'GainStep-RB',
    )

    def __init__(self, devname, props2init='all'):
        """."""
        devname = BunchbyBunch.process_device_name(devname)
        devname = devname.substitute(dev='BbBAmp' + devname.idx, idx='')
        super().__init__(devname, props2init=props2init)

    @property
    def enable(self):
        """."""
        return self['Enbl-Sts']

    @enable.setter
    def enable(self, value):
        self['Enbl-Sel'] = int(value)

    @property
    def auto_gain(self):
        """."""
        return self['GainAuto-Sts']

    @auto_gain.setter
    def auto_gain(self, value):
        self['GainAuto-Sel'] = int(value)

    @property
    def gain(self):
        """."""
        return self['Gain-RB']

    @gain.setter
    def gain(self, value):
        self['Gain-SP'] = value

    @property
    def gain_step(self):
        """."""
        return self['GainStep-RB']

    @gain_step.setter
    def gain_step(self, value):
        self['GainStep-SP'] = value

    def cmd_reset(self, timeout=DEF_TIMEOUT):
        """."""
        self['Rst-Cmd'] = 1
        if not self.wait('Rst-Cmd', 1, timeout / 2):
            return False
        _time.sleep(0.2)
        self['Rst-Cmd'] = 0
        return self.wait('Rst-Cmd', 0, timeout / 2)



"""Define PVs, constants and properties of OrbitCorr SoftIOCs."""
import os as _os
from copy import deepcopy as _dcopy

from .. import csdev as _csdev
from ..diagbeam.bpm.csdev import Const as _CSBPM
from ..namesys import SiriusPVName as _PVName
from ..search import BPMSearch as _BPMSearch, LLTimeSearch as _TISearch, \
    MASearch as _MASearch, PSSearch as _PSSearch
from ..timesys import csdev as _cstiming

# --- Enumeration Types ---


class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    ENBL_RF = _csdev.ETypes.OFF_ON
    OPEN_CLOSED = ('Open', 'Closed')
    ORB_MODE_SI = ('SlowOrb', 'MultiTurn', 'SinglePass')
    ORB_MODE_RINGS = ('MultiTurn', 'SinglePass')
    ORB_MODE_TLINES = ('SinglePass', )
    SMOOTH_METH = ('Average', 'Median')
    RESPMAT_MODE = ('Mxx', 'Myy', 'NoCoup', 'Full')
    SPASS_METHOD = ('FromBPMs', 'Calculated')
    MTURN_ACQUIRE = ('Idle', 'Acquire')
    APPLY_CORR_TLINES = ('CH', 'CV', 'All')
    APPLY_CORR_RINGS = ('CH', 'CV', 'RF', 'All')
    APPLY_DELTA_MON = ('Idle', 'Applying', 'Done', 'Error')
    SI_CORR_SYNC = ('Off', 'Event', 'Clock', 'RmpBO')
    ORB_ACQ_CHAN = ('FAcq', 'FOFB', 'TbT', 'ADC', 'ADCSwp')
    MEAS_RMAT_CMD = ('Start', 'Stop', 'Reset')
    MEAS_RMAT_MON = ('Idle', 'Measuring', 'Completed', 'Aborted')
    DRIVE_TYPE = ('Sine', 'Square', 'Impulse')
    TLINES = ('TB', 'TS')
    RINGS = ('BO', 'SI')
    ACCELERATORS = TLINES + RINGS

    STS_LBLS_CORR_TLINES = (
        "CHCVConnected",
        "CHCVModeConfigured",
        "CHCVPwrStateOn",
    )
    STS_LBLS_CORR_RINGS = STS_LBLS_CORR_TLINES + (
        "TimingConnected",
        "TimingConfigured",
        "RFConnected",
        "RFPwrStateOn",
    )
    STS_LBLS_ORB = (
        "TimingConnected",
        "TimingConfigured",
        "BPMsConnected",
        "BPMsConfigured",
        "BPMsTestDsbld",
        "BPMsSwSyncEnbld",
        "OrbRawConnected",
    )
    STS_LBLS_GLOB = ("Ok", "NotOk")


_et = ETypes  # syntactic sugar


# --- Const class ---


class ConstTLines(_csdev.Const):
    """Const class defining transport lines orbitcorr constants."""

    ORBIT_CONVERSION_UNIT = 1 / 1000  # from nm to um
    MAX_MT_ORBS = 4000
    MAX_DRIVE_DATA = 3 * 5000
    MIN_SING_VAL = 0.2
    TIKHONOV_REG_CONST = 0
    TINY_KICK = 1e-3  # [urad]
    DEF_MAX_ORB_DISTORTION = 50  # [um]
    ACQRATE_TRIGMODE = 2  # [Hz]
    ORBIT_UPDATE_RATE = 60  # [Hz] very high frequency to improve sinc.
    BPMsFreq = 10.48  # [Hz]

    TrigAcqCtrl = _CSBPM.AcqEvents
    TrigAcqChan = _csdev.Const.register("TrigAcqChan", _et.ORB_ACQ_CHAN)
    TrigAcqRepeat = _CSBPM.AcqRepeat
    SmoothMeth = _csdev.Const.register("SmoothMeth", _et.SMOOTH_METH)
    RespMatMode = _csdev.Const.register("RespMatMode", _et.RESPMAT_MODE)
    MeasRespMatCmd = _csdev.Const.register("MeasRespMatCmd", _et.MEAS_RMAT_CMD)
    MeasRespMatMon = _csdev.Const.register("MeasRespMatMon", _et.MEAS_RMAT_MON)
    TransportLines = _csdev.Const.register(
        "TransportLines", _et.TLINES, (0, 1)
    )
    Rings = _csdev.Const.register("Rings", _et.RINGS, (2, 3))
    Accelerators = _csdev.Const.register("Accelerators", _et.ACCELERATORS)

    SOFBMode = _csdev.Const.register("SOFBMode", _et.ORB_MODE_TLINES)
    SyncWithInj = _csdev.Const.register("SyncWithInj", _et.OFF_ON)
    ApplyDelta = _csdev.Const.register("ApplyDelta", _et.APPLY_CORR_TLINES)
    ApplyDeltaMon = _csdev.Const.register("ApplyDeltaMon", _et.APPLY_DELTA_MON)
    StsLblsCorr = _csdev.Const.register(
        "StsLblsCorr", _et.STS_LBLS_CORR_TLINES
    )
    StsLblsOrb = _csdev.Const.register("StsLblsOrb", _et.STS_LBLS_ORB)
    StsLblsGlob = _csdev.Const.register("StsLblsGlob", _et.STS_LBLS_GLOB)

    LoopState = _csdev.Const.register("LoopState", _et.OPEN_CLOSED)


class ConstRings(ConstTLines):
    """Const class defining rings orbitcorr constants."""

    SOFBMode = _csdev.Const.register("SOFBMode", _et.ORB_MODE_RINGS)
    StsLblsCorr = _csdev.Const.register("StsLblsCorr", _et.STS_LBLS_CORR_RINGS)
    MTurnAcquire = _csdev.Const.register("MTurnAcquire", _et.MTURN_ACQUIRE)
    ApplyDelta = _csdev.Const.register("ApplyDelta", _et.APPLY_CORR_RINGS)
    EnblRF = _csdev.Const.register("EnblRF", _et.ENBL_RF)
    RF_GEN_NAME = "RF-Gen"


class ConstSI(ConstRings):
    """Const class defining rings orbitcorr constants."""

    SOFBMode = _csdev.Const.register('SOFBMode', _et.ORB_MODE_SI)
    CorrSync = _csdev.Const.register('CorrSync', _et.SI_CORR_SYNC)
    DriveType = _csdev.Const.register('DriveType', _et.DRIVE_TYPE)
    DriveState = _csdev.Const.register('DriveState', _et.OPEN_CLOSED)

    CORR_DEF_DELAY = 12  # [ms]


# --- Database classes ---


class SOFBTLines(ConstTLines):
    """SOFB class for TLines."""

    def __init__(self, acc):
        """Init1 method."""
        self.acc = acc.upper()
        self.evg_name = _TISearch.get_evg_name()
        self.acc_idx = self.Accelerators._fields.index(self.acc)

        # Define the BPMs:
        self.bpm_names = _BPMSearch.get_names({"sec": acc, "dev": "BPM"})

        # Define correctors:
        filter_ch = dict(sec=acc, dis="PS", dev="CH")
        filter_cv = dict(sec=acc, dis="PS", dev="CV")
        if self.acc == "SI":
            filter_ch.update({"sub": "..(M|C)."})
            filter_cv.update({"sub": "..(M|C)."})
        self.ch_names = _PSSearch.get_psnames(filter_ch)
        self.cv_names = _PSSearch.get_psnames(filter_cv)
        if self.acc == "TS":
            self.ch_names = [_PVName("TS-01:PU-EjeSeptG")] + self.ch_names
            self.cv_names = [
                cvn
                for cvn in self.cv_names
                if not ("E" in cvn.idx or "0" in cvn.idx)
            ]

        # Give them a nickname:
        self.bpm_nicknames = _BPMSearch.get_nicknames(self.bpm_names)
        self.ch_nicknames = _PSSearch.get_psnicknames(self.ch_names)
        self.cv_nicknames = _PSSearch.get_psnicknames(self.cv_names)
        if self.acc == "TS":
            self.ch_nicknames[0] = "EjeseptG"

        # Find their position along the ring:
        self.bpm_pos = _BPMSearch.get_positions(self.bpm_names)
        self.ch_pos = _MASearch.get_mapositions(
            map(
                lambda x: x.substitute(dis="MA" if x.dis == "PS" else "PM"),
                self.ch_names,
            )
        )
        self.cv_pos = _MASearch.get_mapositions(
            map(
                lambda x: x.substitute(dis="MA" if x.dis == "PS" else "PM"),
                self.cv_names,
            )
        )

        # Find the total number of BPMs and correctors:
        self.nr_bpms = len(self.bpm_names)
        self.nr_ch = len(self.ch_names)
        self.nr_cv = len(self.cv_names)
        self.nr_chcv = self.nr_ch + self.nr_cv
        self.nr_corrs = self.nr_chcv + 1 if self.isring else self.nr_chcv

        ext = acc.lower() + "orb"
        ioc_fol = acc.lower() + "-ap-sofb"
        ioc_fol = _os.path.join("/home", "sirius", "iocs-log", ioc_fol, "data")
        self.ref_orb_fname = _os.path.join(ioc_fol, "ref_orbit." + ext)
        ext = acc.lower() + "respmat"
        self.respmat_fname = _os.path.join(ioc_fol, "respmat." + ext)

        self.trigger_acq_name = self.acc + '-Fam:TI-BPM'
        if self.acc == 'SI':
            self.trigger_cor_name = self.acc + '-Glob:TI-Mags-Corrs'
            self.evt_cor_name = 'Orb' + self.acc
            self.clk_cor_name = 'Clock3'
            self.evt_rmpbo = 'RmpBO'

        self.evt_acq_name = "Linac"
        self.matrix_size = self.nr_corrs * (2 * self.nr_bpms)
        self.nr_svals = min(self.nr_corrs, 2 * self.nr_bpms)
        self.circum = 21.2477 if self.acc == "TB" else 26.8933  # in meters
        self.rev_per = self.circum / 299792458  # in seconds

    @property
    def isring(self):
        """."""
        return self.acc in self.Rings._fields

    def get_ioc_database(self, prefix=""):
        """Return IOC database."""
        dbase = self.get_sofb_database(prefix)
        dbase.update(self.get_corrs_database(prefix))
        dbase.update(self.get_respmat_database(prefix))
        dbase.update(self.get_orbit_database(prefix))
        dbase = _csdev.add_pvslist_cte(dbase)
        return dbase

    def get_sofb_database(self, prefix=""):
        """Return SOFB database."""
        dbase = {
            'Log-Mon': {'type': 'char', 'value': '', 'count': 200},
            'MeasRespMat-Cmd': {
                'type': 'enum', 'value': 0,
                'enums': self.MeasRespMatCmd._fields},
            'MeasRespMat-Mon': {
                'type': 'enum', 'value': 0,
                'enums': self.MeasRespMatMon._fields},
            'MeasRespMatKickCH-SP': {
                'type': 'float', 'value': 15, 'unit': 'urad', 'prec': 3,
                'lolim': 0.002, 'hilim': 500},
            'MeasRespMatKickCH-RB': {
                'type': 'float', 'value': 15, 'unit': 'urad', 'prec': 3,
                'lolim': 0.002, 'hilim': 500},
            'MeasRespMatKickCV-SP': {
                'type': 'float', 'value': 22.5, 'unit': 'urad', 'prec': 3,
                'lolim': 0.002, 'hilim': 500},
            'MeasRespMatKickCV-RB': {
                'type': 'float', 'value': 22.5, 'unit': 'urad', 'prec': 3,
                'lolim': 0.002, 'hilim': 500},
            'MeasRespMatWait-SP': {
                'type': 'float', 'value': 1, 'unit': 's', 'prec': 3,
                'lolim': 0.005, 'hilim': 100},
            'MeasRespMatWait-RB': {
                'type': 'float', 'value': 1, 'unit': 's', 'prec': 3,
                'lolim': 0.005, 'hilim': 100},
            'CalcDelta-Cmd': {
                'type': 'int', 'value': 0, 'unit': 'Calculate kicks'},
            'ManCorrGainCH-SP': {
                'type': 'float', 'value': 100, 'unit': '%', 'prec': 2,
                'lolim': -10000, 'hilim': 10000},
            'ManCorrGainCH-RB': {
                'type': 'float', 'value': 100, 'prec': 2, 'unit': '%'},
            'ManCorrGainCV-SP': {
                'type': 'float', 'value': 100, 'unit': '%', 'prec': 2,
                'lolim': -10000, 'hilim': 10000},
            'ManCorrGainCV-RB': {
                'type': 'float', 'value': 100, 'prec': 2, 'unit': '%'},
            'MaxKickCH-SP': {
                'type': 'float', 'value': 300, 'unit': 'urad', 'prec': 3,
                'lolim': 0, 'hilim': 500000},
            'MaxKickCH-RB': {
                'type': 'float', 'value': 300, 'prec': 3, 'unit': 'urad',
                'lolim': 0, 'hilim': 500000},
            'MaxKickCV-SP': {
                'type': 'float', 'value': 300, 'unit': 'urad', 'prec': 3,
                'lolim': 0, 'hilim': 10000},
            'MaxKickCV-RB': {
                'type': 'float', 'value': 300, 'prec': 3, 'unit': 'urad',
                'lolim': 0, 'hilim': 10000},
            'MaxDeltaKickCH-SP': {
                'type': 'float', 'value': 5, 'unit': 'urad', 'prec': 3,
                'lolim': 0, 'hilim': 10000},
            'MaxDeltaKickCH-RB': {
                'type': 'float', 'value': 5, 'prec': 3, 'unit': 'urad',
                'lolim': 0, 'hilim': 10000},
            'MaxDeltaKickCV-SP': {
                'type': 'float', 'value': 5, 'unit': 'urad', 'prec': 3,
                'lolim': 0, 'hilim': 10000},
            'MaxDeltaKickCV-RB': {
                'type': 'float', 'value': 5, 'prec': 3, 'unit': 'urad',
                'lolim': 0, 'hilim': 10000},
            'DeltaKickCH-SP': {
                'type': 'float', 'count': self.nr_ch, 'value': self.nr_ch*[0],
                'unit': 'urad'},
            'DeltaKickCH-RB': {
                'type': 'float', 'count': self.nr_ch, 'value': self.nr_ch*[0],
                'unit': 'urad'},
            'DeltaKickCV-SP': {
                'type': 'float', 'count': self.nr_cv, 'value': self.nr_cv*[0],
                'unit': 'urad'},
            'DeltaKickCV-RB': {
                'type': 'float', 'count': self.nr_cv, 'value': self.nr_cv*[0],
                'unit': 'urad'},
            'ApplyDelta-Cmd': {
                'type': 'enum', 'enums': self.ApplyDelta._fields, 'value': 0,
                'unit': 'Apply last calculated kicks.'},
            'ApplyDelta-Mon': {
                'type': 'enum', 'enums': self.ApplyDeltaMon._fields,
                'value': 0, 'unit': 'Status of Kicks implementation.'},
            'Status-Mon': {
                'type': 'enum', 'value': 1,
                'enums': self.StsLblsGlob._fields},
            'RefKickCH-Mon': {
                'type': 'float',
                'count': self.nr_ch,
                'value': self.nr_ch*[0],
                'unit': 'urad'
            },
            'RefKickCV-Mon': {
                'type': 'float',
                'count': self.nr_cv,
                'value': self.nr_cv*[0],
                'unit': 'urad'
            },
        }
        return self._add_prefix(dbase, prefix)

    def get_corrs_database(self, prefix=""):
        """Return SOFB Correctors database."""
        dbase = {
            "KickCH-Mon": {
                "type": "float",
                "count": self.nr_ch,
                "value": self.nr_ch * [0],
                "unit": "urad",
            },
            "KickCV-Mon": {
                "type": "float",
                "count": self.nr_cv,
                "value": self.nr_cv * [0],
                "unit": "urad",
            },
            "CorrConfig-Cmd": {"type": "int", "value": 0},
            "CHPosS-Cte": {
                "type": "float",
                "unit": "m",
                "count": self.nr_ch,
                "value": self.ch_pos,
            },
            "CVPosS-Cte": {
                "type": "float",
                "unit": "m",
                "count": self.nr_cv,
                "value": self.cv_pos,
            },
            "CHNickName-Cte": {
                "type": "string",
                "unit": "shortname for the chs.",
                "count": self.nr_ch,
                "value": self.ch_nicknames,
            },
            "CVNickName-Cte": {
                "type": "string",
                "unit": "shortname for the cvs.",
                "count": self.nr_cv,
                "value": self.cv_nicknames,
            },
            "CorrStatus-Mon": {"type": "int", "value": 0b1111111},
            "CorrStatusLabels-Cte": {
                "type": "string",
                "count": len(self.StsLblsCorr._fields),
                "value": self.StsLblsCorr._fields,
            },
        }
        return self._add_prefix(dbase, prefix)

    def get_orbit_database(self, prefix=""):
        """Return Orbit database."""
        nbpm = self.nr_bpms
        pvs = [
            "RefOrbX-SP",
            "RefOrbX-RB",
            "RefOrbY-SP",
            "RefOrbY-RB",
            "SPassOrbX-Mon",
            "SPassOrbY-Mon",
            "SPassSum-Mon",
            "BPMOffsetX-Mon",
            "BPMOffsetY-Mon",
        ]
        dbase = dict()
        prop = {
            "type": "float",
            "unit": "um",
            "count": nbpm,
            "value": nbpm * [0],
        }
        for k in pvs:
            dbase[k] = _dcopy(prop)

        # Orbit statistics
        pvs = [
            "DeltaOrbXAvg-Mon",
            "DeltaOrbYAvg-Mon",
            "DeltaOrbXStd-Mon",
            "DeltaOrbYStd-Mon",
            "DeltaOrbXMin-Mon",
            "DeltaOrbYMin-Mon",
            "DeltaOrbXMax-Mon",
            "DeltaOrbYMax-Mon",
        ]
        prop = {"type": "float", "value": 0, "prec": 3, "unit": "um"}
        for k in pvs:
            dbase[k] = _dcopy(prop)

        dbase.update(
            {
                "SyncWithInjection-Sel": {
                    "type": "enum",
                    "unit": "Off_On",
                    "value": self.SyncWithInj.Off,
                    "enums": self.SyncWithInj._fields,
                },
                "SyncWithInjection-Sts": {
                    "type": "enum",
                    "unit": "Off_On",
                    "value": self.SyncWithInj.Off,
                    "enums": self.SyncWithInj._fields,
                },
                "TestDataEnbl-Sel": {
                    "type": "enum",
                    "unit": "Dsbl_Enbl",
                    "value": self.DsblEnbl.Dsbl,
                    "enums": self.DsblEnbl._fields,
                },
                "TestDataEnbl-Sts": {
                    "type": "enum",
                    "unit": "Dsbl_Enbl",
                    "value": self.DsblEnbl.Dsbl,
                    "enums": self.DsblEnbl._fields,
                },
                "TrigAcqConfig-Cmd": {"type": "int", "value": 0},
                "TrigAcqCtrl-Sel": {
                    "type": "enum",
                    "unit": "Start_Stop",
                    "value": self.TrigAcqCtrl.Stop,
                    "enums": self.TrigAcqCtrl._fields,
                },
                "TrigAcqCtrl-Sts": {
                    "type": "enum",
                    "unit": "Start_Stop.",
                    "value": self.TrigAcqCtrl.Stop,
                    "enums": self.TrigAcqCtrl._fields,
                },
                "TrigAcqChan-Sel": {
                    "type": "enum",
                    "unit": "FAcq_FOFB_TbT_ADC_ADCSwp",
                    "value": self.TrigAcqChan.ADCSwp,
                    "enums": self.TrigAcqChan._fields,
                },
                "TrigAcqChan-Sts": {
                    "type": "enum",
                    "unit": "FAcq_FOFB_TbT_ADC_ADCSwp",
                    "value": self.TrigAcqChan.ADCSwp,
                    "enums": self.TrigAcqChan._fields,
                },
                "TrigAcqRepeat-Sel": {
                    "type": "enum",
                    "unit": "Normal_Repetitive",
                    "value": self.TrigAcqRepeat.Repetitive,
                    "enums": self.TrigAcqRepeat._fields,
                },
                "TrigAcqRepeat-Sts": {
                    "type": "enum",
                    "unit": "Normal_Repetitive",
                    "value": self.TrigAcqRepeat.Repetitive,
                    "enums": self.TrigAcqRepeat._fields,
                },
                "TrigNrShots-SP": {
                    "type": "int",
                    "unit": "",
                    "value": 1,
                    "hilim": 1000,
                    "lolim": 1,
                },
                "TrigNrShots-RB": {
                    "type": "int",
                    "unit": "",
                    "value": 1,
                    "hilim": 1000,
                    "lolim": 1,
                },
                "TrigNrSamplesPre-SP": {
                    "type": "int",
                    "unit": "",
                    "value": 0,
                    "hilim": 20000,
                    "lolim": -1,
                },
                "TrigNrSamplesPre-RB": {
                    "type": "int",
                    "unit": "",
                    "value": 0,
                    "hilim": 20000,
                    "lolim": -1,
                },
                "TrigNrSamplesPost-SP": {
                    "type": "int",
                    "unit": "",
                    "value": 382,
                    "hilim": 20000,
                    "lolim": 0,
                },
                "TrigNrSamplesPost-RB": {
                    "type": "int",
                    "unit": "",
                    "value": 382,
                    "hilim": 20000,
                    "lolim": 0,
                },
                "PolyCalibration-Sel": {
                    "type": "enum",
                    "unit": "Dsbl_Enbl",
                    "value": self.DsblEnbl.Enbl,
                    "enums": self.DsblEnbl._fields,
                },
                "PolyCalibration-Sts": {
                    "type": "enum",
                    "unit": "Dsbl_Enbl",
                    "value": self.DsblEnbl.Enbl,
                    "enums": self.DsblEnbl._fields,
                },
                "SmoothNrPts-SP": {
                    "type": "int",
                    "value": 1,
                    "unit": "#",
                    "lolim": 1,
                    "hilim": 500,
                },
                "SmoothNrPts-RB": {
                    "type": "int",
                    "value": 1,
                    "unit": "#",
                    "lolim": 1,
                    "hilim": 500,
                },
                "SmoothMethod-Sel": {
                    "type": "enum",
                    "unit": "Average_Median",
                    "value": self.SmoothMeth.Average,
                    "enums": _et.SMOOTH_METH,
                },
                "SmoothMethod-Sts": {
                    "type": "enum",
                    "unit": "Average_Median",
                    "value": self.SmoothMeth.Average,
                    "enums": _et.SMOOTH_METH,
                },
                "SmoothReset-Cmd": {"type": "int", "value": 0, "unit": "#"},
                "BufferCount-Mon": {"type": "int", "value": 0, "unit": "#"},
                "SPassMaskSplBeg-SP": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1000,
                },
                "SPassMaskSplBeg-RB": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1000,
                },
                "SPassMaskSplEnd-SP": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1000,
                },
                "SPassMaskSplEnd-RB": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1000,
                },
                "SPassAvgNrTurns-SP": {
                    "type": "int",
                    "value": 1,
                    "lolim": 1,
                    "hilim": 1000,
                },
                "SPassAvgNrTurns-RB": {
                    "type": "int",
                    "value": 1,
                    "lolim": 1,
                    "hilim": 1000,
                },
                "BPMPosS-Mon": {
                    "type": "float",
                    "unit": "m",
                    "count": nbpm,
                    "value": self.bpm_pos,
                    "prec": 2,
                },
                "BPMNickName-Cte": {
                    "type": "string",
                    "unit": "shortname for the bpms.",
                    "count": nbpm,
                    "value": self.bpm_nicknames,
                },
                "OrbStatus-Mon": {"type": "int", "value": 0b00000000},
                "OrbStatusLabels-Cte": {
                    "type": "string",
                    "count": len(self.StsLblsOrb._fields),
                    "value": self.StsLblsOrb._fields,
                },
                "SlowOrbTimeout-Mon": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1001,
                },
                "SyncBPMs-Cmd": {"type": "int", "value": 0},
            }
        )
        return self._add_prefix(dbase, prefix)

    def get_respmat_database(self, prefix=""):
        """Return SOFB respmat database."""
        dbase = {
            "RespMat-SP": {
                "type": "float",
                "count": self.matrix_size,
                "value": self.matrix_size * [0],
                "unit": "(BH, BV)(um) x (CH, CV, RF)(urad, Hz)",
            },
            "RespMat-RB": {
                "type": "float",
                "count": self.matrix_size,
                "value": self.matrix_size * [0],
                "unit": "(BH, BV)(um) x (CH, CV, RF)(urad, Hz)",
            },
            "RespMat-Mon": {
                "type": "float",
                "count": self.matrix_size,
                "value": self.matrix_size * [0],
                "unit": "(BH, BV)(um) x (CH, CV, RF)(urad, Hz)",
            },
            "InvRespMat-Mon": {
                "type": "float",
                "count": self.matrix_size,
                "value": self.matrix_size * [0],
                "unit": "(CH, CV, RF)(urad, Hz) x (BH, BV)(um)",
            },
            "RespMatMode-Sel": {
                "type": "enum",
                "value": self.RespMatMode.Full,
                "enums": self.RespMatMode._fields,
            },
            "RespMatMode-Sts": {
                "type": "enum",
                "value": self.RespMatMode.Full,
                "enums": self.RespMatMode._fields,
            },
            "SingValuesRaw-Mon": {
                "type": "float",
                "count": self.nr_svals,
                "value": self.nr_svals * [0],
                "unit": "Singular values of the matrix",
            },
            "SingValues-Mon": {
                "type": "float",
                "count": self.nr_svals,
                "value": self.nr_svals * [0],
                "unit": "Singular values of the matrix in use",
            },
            "CHEnblList-SP": {
                "type": "int",
                "count": self.nr_ch,
                "value": self.nr_ch * [1],
                "unit": "CHs used in correction",
            },
            "CHEnblList-RB": {
                "type": "int",
                "count": self.nr_ch,
                "value": self.nr_ch * [1],
                "unit": "CHs used in correction",
            },
            "CVEnblList-SP": {
                "type": "int",
                "count": self.nr_cv,
                "value": self.nr_cv * [1],
                "unit": "CVs used in correction",
            },
            "CVEnblList-RB": {
                "type": "int",
                "count": self.nr_cv,
                "value": self.nr_cv * [1],
                "unit": "CVs used in correction",
            },
            "BPMXEnblList-SP": {
                "type": "int",
                "count": self.nr_bpms,
                "value": self.nr_bpms * [1],
                "unit": "BPMX used in correction",
            },
            "BPMXEnblList-RB": {
                "type": "int",
                "count": self.nr_bpms,
                "value": self.nr_bpms * [1],
                "unit": "BPMX used in correction",
            },
            "BPMYEnblList-SP": {
                "type": "int",
                "count": self.nr_bpms,
                "value": self.nr_bpms * [1],
                "unit": "BPMY used in correction",
            },
            "BPMYEnblList-RB": {
                "type": "int",
                "count": self.nr_bpms,
                "value": self.nr_bpms * [1],
                "unit": "BPMY used in correction",
            },
            "MinSingValue-SP": {
                "type": "float",
                "value": self.MIN_SING_VAL,
                "prec": 5,
                "lolim": 0,
                "hilim": 1e9,
                "unit": "SingVal",
            },
            "MinSingValue-RB": {
                "type": "float",
                "value": self.MIN_SING_VAL,
                "prec": 5,
                "lolim": 0,
                "hilim": 1e9,
                "unit": "SingVal",
            },
            "TikhonovRegConst-SP": {
                "type": "float",
                "value": self.TIKHONOV_REG_CONST,
                "prec": 5,
                "lolim": 0,
                "hilim": 1e9,
                "unit": "SingVal",
            },
            "TikhonovRegConst-RB": {
                "type": "float",
                "value": self.TIKHONOV_REG_CONST,
                "prec": 5,
                "lolim": 0,
                "hilim": 1e9,
                "unit": "SingVal",
            },
            "NrSingValues-Mon": {
                "type": "int",
                "value": self.nr_svals,
                "unit": "#SingVals",
                "lolim": 1,
                "hilim": self.nr_svals,
            },
            "DeltaKickCH-Mon": {
                "type": "float",
                "count": self.nr_ch,
                "value": self.nr_ch * [0],
                "unit": "urad",
            },
            "DeltaKickCV-Mon": {
                "type": "float",
                "count": self.nr_cv,
                "value": self.nr_cv * [0],
                "unit": "urad",
            },
        }
        return self._add_prefix(dbase, prefix)

    def _add_prefix(self, dbase, prefix):
        if prefix:
            return {prefix + k: v for k, v in dbase.items()}
        return dbase


class SOFBRings(SOFBTLines, ConstRings):
    """SOFB class."""

    def __init__(self, acc):
        """Init method."""
        SOFBTLines.__init__(self, acc)
        self.circum = 496.8  # in meter
        self.harm_number = 828
        self.rev_per = self.circum / 299792458  # in seconds

    def get_sofb_database(self, prefix=""):
        """Return SOFB database."""
        db_ring = {
            'MeasRespMatKickRF-SP': {
                'type': 'float', 'value': 75, 'unit': 'Hz', 'prec': 2,
                'lolim': 1, 'hilim': 1000},
            'MeasRespMatKickRF-RB': {
                'type': 'float', 'value': 75, 'unit': 'Hz', 'prec': 2,
                'lolim': 1, 'hilim': 1000},
            'ManCorrGainRF-SP': {
                'type': 'float', 'value': 100, 'unit': '%', 'prec': 2,
                'lolim': -1000, 'hilim': 1000},
            'ManCorrGainRF-RB': {
                'type': 'float', 'value': 100, 'prec': 2, 'unit': '%'},
            'MaxDeltaKickRF-SP': {
                'type': 'float', 'value': 10, 'unit': 'Hz', 'prec': 2,
                'lolim': 0, 'hilim': 10000},
            'MaxDeltaKickRF-RB': {
                'type': 'float', 'value': 10, 'prec': 2, 'unit': 'Hz',
                'lolim': 0, 'hilim': 10000},
            'DeltaKickRF-SP': {
                'type': 'float', 'value': 0, 'prec': 2, 'unit': 'Hz'},
            'DeltaKickRF-RB': {
                'type': 'float', 'value': 0, 'prec': 2, 'unit': 'Hz'},
            'RefKickRF-Mon': {
                'type': 'float', 'value': 0, 'prec': 2, 'unit': 'Hz',
            },
        }
        dbase = super().get_sofb_database(prefix=prefix)
        dbase.update(self._add_prefix(db_ring, prefix))
        return dbase

    def get_corrs_database(self, prefix=""):
        """Return SOFB correctors database."""
        db_ring = {
            "KickRF-Mon": {
                "type": "float",
                "value": 1,
                "unit": "Hz",
                "prec": 2,
            },
            "OrbLength-Mon": {
                "type": "float",
                "value": 1,
                "unit": "m",
                "prec": 6,
            },
        }
        dbase = super().get_corrs_database(prefix=prefix)
        dbase.update(self._add_prefix(db_ring, prefix))
        return dbase

    def get_respmat_database(self, prefix=""):
        """Return SOFB respmat database."""
        db_ring = {
            "RFEnbl-Sel": {
                "type": "enum",
                "enums": self.EnblRF._fields,
                "value": 1,
                "unit": "Off_On",
            },
            "RFEnbl-Sts": {
                "type": "enum",
                "enums": self.EnblRF._fields,
                "value": 1,
                "unit": "Off_On",
            },
            "DeltaKickRF-Mon": {
                "type": "float",
                "value": 0,
                "prec": 2,
                "unit": "Hz",
            },
        }
        dbase = super().get_respmat_database(prefix=prefix)
        dbase.update(self._add_prefix(db_ring, prefix))
        return dbase

    def get_orbit_database(self, prefix=""):
        """Return Orbit database."""
        nbpm = self.nr_bpms
        pvs_ring = ["MTurnIdxOrbX-Mon", "MTurnIdxOrbY-Mon", "MTurnIdxSum-Mon"]
        db_ring = dict()
        prop = {
            "type": "float",
            "unit": "um",
            "count": nbpm,
            "value": nbpm * [0],
        }
        for k in pvs_ring:
            db_ring[k] = _dcopy(prop)
        db_ring.update(
            {
                "MTurnAcquire-Cmd": {
                    "type": "enum",
                    "value": 0,
                    "enums": self.MTurnAcquire._fields,
                },
                "MTurnUseMask-Sel": {
                    "type": "enum",
                    "value": self.DsblEnbl.Dsbl,
                    "enums": self.DsblEnbl._fields,
                },
                "MTurnUseMask-Sts": {
                    "type": "enum",
                    "value": self.DsblEnbl.Dsbl,
                    "enums": self.DsblEnbl._fields,
                },
                "MTurnMaskSplBeg-SP": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1000,
                },
                "MTurnMaskSplBeg-RB": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1000,
                },
                "MTurnMaskSplEnd-SP": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1000,
                },
                "MTurnMaskSplEnd-RB": {
                    "type": "int",
                    "value": 0,
                    "lolim": -1,
                    "hilim": 1000,
                },
                "MTurnDownSample-SP": {
                    "type": "int",
                    "unit": "",
                    "value": 1,
                    "hilim": 20000,
                    "lolim": 1,
                },
                "MTurnDownSample-RB": {
                    "type": "int",
                    "unit": "",
                    "value": 1,
                    "hilim": 20000,
                    "lolim": 1,
                },
                "MTurnOrbX-Mon": {
                    "type": "float",
                    "unit": "um",
                    "count": self.MAX_MT_ORBS * nbpm,
                    "value": 50 * nbpm * [0],
                },
                "MTurnOrbY-Mon": {
                    "type": "float",
                    "unit": "um",
                    "count": self.MAX_MT_ORBS * nbpm,
                    "value": 50 * nbpm * [0],
                },
                "MTurnSum-Mon": {
                    "type": "float",
                    "unit": "count",
                    "count": self.MAX_MT_ORBS * nbpm,
                    "value": 50 * nbpm * [0],
                },
                "MTurnTime-Mon": {
                    "type": "float",
                    "unit": "s",
                    "count": self.MAX_MT_ORBS,
                    "value": 50 * [0],
                },
                "MTurnIdx-SP": {
                    "type": "int",
                    "unit": "",
                    "value": 0,
                    "hilim": self.MAX_MT_ORBS,
                    "lolim": 0,
                },
                "MTurnIdx-RB": {
                    "type": "int",
                    "unit": "",
                    "value": 0,
                    "hilim": self.MAX_MT_ORBS,
                    "lolim": 0,
                },
                "MTurnIdxTime-Mon": {
                    "type": "float",
                    "unit": "s",
                    "value": 0.0,
                    "prec": 5,
                    "hilim": 500,
                    "lolim": 0,
                },
            }
        )
        dbase = super().get_orbit_database(prefix=prefix)
        dbase['TrigAcqChan-Sel']['value'] = self.TrigAcqChan.TbT
        dbase['TrigAcqChan-Sts']['value'] = self.TrigAcqChan.TbT
        dbase.update(self._add_prefix(db_ring, prefix))
        return dbase


class SOFBSI(SOFBRings, ConstSI):
    """SOFB class."""

    def __init__(self, acc):
        """Init method."""
        SOFBRings.__init__(self, acc)
        vals = _cstiming.get_hl_trigger_database(self.trigger_cor_name)
        evts = vals["Src-Sel"]["enums"]
        self.CorrExtEvtSrc = self.register("CorrExtEvtSrc", evts)
        self.circum = 518.396  # in meter
        self.harm_number = 864
        self.rev_per = self.circum / 299792458  # in seconds

    def get_sofb_database(self, prefix=""):
        """Return SOFB database."""
        db_ring = {
            'LoopState-Sel': {
                'type': 'enum', 'enums': self.LoopState._fields, 'value': 0},
            'LoopState-Sts': {
                'type': 'enum', 'enums': self.LoopState._fields, 'value': 0},
            'LoopFreq-SP': {
                'type': 'float', 'value': self.BPMsFreq / 10, 'unit': 'Hz',
                'prec': 3, 'lolim': 1e-3, 'hilim': 60},
            'LoopFreq-RB': {
                'type': 'float', 'value': self.BPMsFreq / 10, 'unit': 'Hz',
                'prec': 3, 'lolim': 1e-3, 'hilim': 60},
            'LoopPIDKpCH-SP': {
                'type': 'float', 'value': 0.0, 'unit': 'frac', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKpCH-RB': {
                'type': 'float', 'value': 0.0, 'unit': 'frac', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKpCV-SP': {
                'type': 'float', 'value': 0.0, 'unit': 'frac', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKpCV-RB': {
                'type': 'float', 'value': 0.0, 'unit': 'frac', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKiCH-SP': {
                'type': 'float', 'value': 0.2, 'unit': 'frac.Hz', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKiCH-RB': {
                'type': 'float', 'value': 0.2, 'unit': 'frac.Hz', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKiCV-SP': {
                'type': 'float', 'value': 0.2, 'unit': 'frac.Hz', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKiCV-RB': {
                'type': 'float', 'value': 0.2, 'unit': 'frac.Hz', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKdCH-SP': {
                'type': 'float', 'value': 0, 'unit': 'frac.s', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKdCH-RB': {
                'type': 'float', 'value': 0, 'unit': 'frac.s', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKdCV-SP': {
                'type': 'float', 'value': 0, 'unit': 'frac.s', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKdCV-RB': {
                'type': 'float', 'value': 0, 'unit': 'frac.s', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKpRF-SP': {
                'type': 'float', 'value': 0.0, 'unit': 'frac', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKpRF-RB': {
                'type': 'float', 'value': 0.0, 'unit': 'frac', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKiRF-SP': {
                'type': 'float', 'value': 0.2, 'unit': 'frac.Hz', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKiRF-RB': {
                'type': 'float', 'value': 0.2, 'unit': 'frac.Hz', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKdRF-SP': {
                'type': 'float', 'value': 0, 'unit': 'frac.s', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPIDKdRF-RB': {
                'type': 'float', 'value': 0, 'unit': 'frac.s', 'prec': 3,
                'lolim': -1000, 'hilim': 1000},
            'LoopPerfItersOk-Mon': {
                'type': 'float', 'value': 0, 'unit': '%', 'prec': 3,
                'lolim': -1, 'hilim': 100},
            'LoopEffectiveRate-Mon': {
                'type': 'float', 'value': 0, 'unit': 'Hz', 'prec': 3,
                'lolim': 0, 'hilim': 100},
            'LoopNumIters-Mon': {
                'type': 'float', 'value': 0, 'unit': '#', 'prec': 0,
                'lolim': 0, 'hilim': 100000},
            'LoopPrintEveryNumIters-SP': {
                'type': 'float', 'value': 200, 'unit': '#', 'prec': 0,
                'lolim': 1, 'hilim': 100000},
            'LoopPrintEveryNumIters-RB': {
                'type': 'float', 'value': 200, 'unit': '#', 'prec': 0,
                'lolim': 1, 'hilim': 100000},
            'LoopPerfItersTOut-Mon': {
                'type': 'float', 'value': 0, 'unit': '%', 'prec': 3,
                'lolim': -1, 'hilim': 100},
            'LoopPerfItersDiff-Mon': {
                'type': 'float', 'value': 0, 'unit': '%', 'prec': 3,
                'lolim': -1, 'hilim': 100},
            'LoopPerfDiffNrPSMax-Mon': {
                'type': 'float', 'value': 0, 'unit': '#', 'prec': 3,
                'lolim': -1, 'hilim': 400},
            'LoopPerfDiffNrPSAvg-Mon': {
                'type': 'float', 'value': 0, 'unit': '#', 'prec': 3,
                'lolim': -1, 'hilim': 400},
            'LoopPerfDiffNrPSStd-Mon': {
                'type': 'float', 'value': 0, 'unit': '#', 'prec': 3,
                'lolim': -1, 'hilim': 400},
            'LoopPerfTimGetOMax-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimGetOMin-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimGetOAvg-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimGetOStd-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimCalcMax-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimCalcMin-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimCalcAvg-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimCalcStd-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimProcMax-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimProcMin-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimProcAvg-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimProcStd-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimAppMax-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimAppMin-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimAppAvg-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimAppStd-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimTotMax-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimTotMin-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimTotAvg-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopPerfTimTotStd-Mon': {
                'type': 'float', 'value': 0, 'unit': 'ms', 'prec': 1,
                'lolim': -1, 'hilim': 100},
            'LoopMaxOrbDistortion-SP': {
                'type': 'float', 'value': self.DEF_MAX_ORB_DISTORTION,
                'prec': 3, 'unit': 'um',
                'lolim': 0, 'hilim': 10000},
            'LoopMaxOrbDistortion-RB': {
                'type': 'float', 'value': self.DEF_MAX_ORB_DISTORTION,
                'prec': 3, 'unit': 'um',
                'lolim': 0, 'hilim': 10000},
            'FOFBDownloadKicksPerc-SP': {
                'type': 'float', 'value': 40.0, 'prec': 2, 'unit': '%',
                'lolim': 0.0, 'hilim': 300.1},
            'FOFBDownloadKicksPerc-RB': {
                'type': 'float', 'value': 40.0, 'prec': 2, 'unit': '%',
                'lolim': 0.0, 'hilim': 300.1},
            'FOFBDownloadKicks-Sel': {
                'type': 'enum', 'value': self.DsblEnbl.Enbl,
                'enums': self.DsblEnbl._fields},
            'FOFBDownloadKicks-Sts': {
                'type': 'enum', 'value': self.DsblEnbl.Enbl,
                'enums': self.DsblEnbl._fields},
            'FOFBDownloadKicks-Mon': {
                'type': 'enum', 'value': self.DsblEnbl.Dsbl,
                'enums': self.DsblEnbl._fields},
            'FOFBDownloadKicksCH-Mon': {
                'type': 'float',
                'count': self.nr_ch,
                'value': self.nr_ch*[0],
                'unit': 'urad'
            },
            'FOFBDownloadKicksCV-Mon': {
                'type': 'float',
                'count': self.nr_cv,
                'value': self.nr_cv*[0],
                'unit': 'urad'
            },
            'FOFBDownloadKicksRF-Mon': {
                'type': 'float', 'value': 0, 'prec': 3, 'unit': 'Hz',
            },
            'FOFBUpdateRefOrbRate-SP': {
                'type': 'int', 'value': 3, 'unit': 'LoopFreqDiv',
                'lolim': 1, 'hilim': 100,
            },
            'FOFBUpdateRefOrbRate-RB': {
                'type': 'int', 'value': 3, 'unit': 'LoopFreqDiv',
                'lolim': 1, 'hilim': 100,
            },
            'FOFBUpdateRefOrbPerc-SP': {
                'type': 'float', 'value': 100.0, 'prec': 2, 'unit': '%',
                'lolim': -0.1, 'hilim': 100.1},
            'FOFBUpdateRefOrbPerc-RB': {
                'type': 'float', 'value': 100.0, 'prec': 2, 'unit': '%',
                'lolim': -0.1, 'hilim': 100.1},
            'FOFBUpdateRefOrb-Sel': {
                'type': 'enum', 'value': self.DsblEnbl.Enbl,
                'enums': self.DsblEnbl._fields},
            'FOFBUpdateRefOrb-Sts': {
                'type': 'enum', 'value': self.DsblEnbl.Enbl,
                'enums': self.DsblEnbl._fields},
            'FOFBUpdateRefOrb-Mon': {
                'type': 'enum', 'value': self.DsblEnbl.Dsbl,
                'enums': self.DsblEnbl._fields},
            'FOFBDeltaRefOrbX-Mon': {
                'type': 'float',
                'unit': 'um',
                'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
            },
            'FOFBDeltaRefOrbY-Mon': {
                'type': 'float',
                'unit': 'um',
                'count': self.nr_bpms,
                'value': self.nr_bpms*[0],
            },
            'DriveFreqDivisor-SP': {
                'type': 'int', 'value': 12, 'unit': 'Div',
                'lolim': 0, 'hilim': 1000},
            'DriveFreqDivisor-RB': {
                'type': 'int', 'value': 12, 'unit': 'Div',
                'lolim': 0, 'hilim': 1000},
            'DriveFrequency-Mon': {
                'type': 'float', 'value': self.BPMsFreq/12, 'prec': 3,
                'unit': 'Hz', 'lolim': 0, 'hilim': 1000},
            'DriveNrCycles-SP': {
                'type': 'int', 'value': 10, 'unit': 'number',
                'lolim': 0, 'hilim': 1000},
            'DriveNrCycles-RB': {
                'type': 'int', 'value': 10, 'unit': 'number',
                'lolim': 0, 'hilim': 1000},
            'DriveDuration-Mon': {
                'type': 'float', 'value': 12/self.BPMsFreq*10, 'prec': 1,
                'unit': 's', 'lolim': 0, 'hilim': 1000},
            'DriveAmplitude-SP': {
                'type': 'float', 'value': 5, 'prec': 2, 'unit': 'urad or Hz',
                'lolim': -100, 'hilim': 100},
            'DriveAmplitude-RB': {
                'type': 'float', 'value': 5, 'prec': 2, 'unit': 'urad or Hz',
                'lolim': -100, 'hilim': 100},
            'DrivePhase-SP': {
                'type': 'float', 'value': 0, 'prec': 3, 'unit': 'deg',
                'lolim': -360, 'hilim': 360},
            'DrivePhase-RB': {
                'type': 'float', 'value': 0, 'prec': 3, 'unit': 'deg',
                'lolim': -360, 'hilim': 360},
            'DriveCorrIndex-SP': {
                'type': 'int', 'value': 0, 'unit': '#',
                'lolim': -self.nr_corrs, 'hilim': self.nr_corrs},
            'DriveCorrIndex-RB': {
                'type': 'int', 'value': 0, 'unit': '#',
                'lolim': -self.nr_corrs, 'hilim': self.nr_corrs},
            'DriveBPMIndex-SP': {
                'type': 'int', 'value': 0, 'unit': '#',
                'lolim': -self.nr_bpms*2, 'hilim': self.nr_bpms*2},
            'DriveBPMIndex-RB': {
                'type': 'int', 'value': 0, 'unit': '#',
                'lolim': -self.nr_bpms*2, 'hilim': self.nr_bpms*2},
            'DriveType-Sel': {
                'type': 'enum', 'enums': self.DriveType._fields,
                'unit': 'Sine_Square_Impulse', 'value': 0},
            'DriveType-Sts': {
                'type': 'enum', 'enums': self.DriveType._fields,
                'unit': 'Sine_Square_Impulse', 'value': 0},
            'DriveState-Sel': {
                'type': 'enum', 'enums': self.DriveState._fields,
                'unit': 'Open_Closed', 'value': 0},
            'DriveState-Sts': {
                'type': 'enum', 'enums': self.DriveState._fields,
                'unit': 'Open_Closed', 'value': 0},
            'DriveData-Mon': {
                'type': 'float', 'unit': '(s, urad, um)',
                'count': self.MAX_DRIVE_DATA, 'value': self.MAX_DRIVE_DATA*[0]}
            }
        dbase = super().get_sofb_database(prefix=prefix)
        dbase.update(self._add_prefix(db_ring, prefix))
        return dbase

    def get_corrs_database(self, prefix=""):
        """Return SOFB correctors database."""
        db_ring = {
            'CorrSync-Sel': {
                'type': 'enum', 'enums': self.CorrSync._fields,
                'unit': 'Off_Event_Clock_RmpBO', 'value': self.CorrSync.Off},
            'CorrSync-Sts': {
                'type': 'enum', 'enums': self.CorrSync._fields,
                'unit': 'Off_Event_Clock_RmpBO', 'value': self.CorrSync.Off},
            }
        dbase = super().get_corrs_database(prefix=prefix)
        dbase.update(self._add_prefix(db_ring, prefix))
        return dbase

    def get_orbit_database(self, prefix=""):
        """Return Orbit database."""
        nbpm = self.nr_bpms
        pvs_ring = ["SlowOrbX-Mon", "SlowOrbY-Mon"]
        db_ring = dict()
        prop = {
            "type": "float",
            "unit": "um",
            "count": nbpm,
            "value": nbpm * [0],
        }
        for k in pvs_ring:
            db_ring[k] = _dcopy(prop)
        db_ring.update(
            {
                "SOFBMode-Sel": {
                    "type": "enum",
                    "unit": "SlowOrb_MultiTurn_SinglePass",
                    "value": 0,
                    "enums": self.SOFBMode._fields,
                },
                "SOFBMode-Sts": {
                    "type": "enum",
                    "unit": "SlowOrb_MultiTurn_SinglePass",
                    "value": 0,
                    "enums": self.SOFBMode._fields,
                },
            }
        )
        dbase = super().get_orbit_database(prefix=prefix)
        dbase["SmoothNrPts-SP"]['value'] = 20
        dbase["SmoothNrPts-RB"]['value'] = 20
        dbase.update(self._add_prefix(db_ring, prefix))
        return dbase


class SOFBFactory:
    """Factory class for SOFB objects."""

    @staticmethod
    def create(acc):
        """Return appropriate SOFB object."""
        acc = acc.upper()
        if acc == "SI":
            return SOFBSI(acc)
        elif acc in _et.RINGS:
            return SOFBRings(acc)
        elif acc in _et.TLINES:
            return SOFBTLines(acc)
        else:
            raise ValueError('Invalid accelerator name "{}"'.format(acc))



"""Subpackage for the Archiver server."""

from ..envars import (
    SRVURL_ARCHIVER as SERVER_URL,
    SRVURL_ARCHIVER_OFFLINE_DATA as SERVER_OFFLINE_URL
)
from . import exceptions
from .client import ClientArchiver
from .devices import Correctors, Orbit, TrimQuads
from .pvarch import PVData, PVDataSet, PVDetails
from .time import Time

del client, pvarch, devices



"""Module to deal with orbit acquisition."""
import logging as _log
import os as _os
import time as _time
import traceback as _traceback
from functools import partial as _part
from threading import Event as _Event, Lock

import bottleneck as _bn
import numpy as _np

from .. import util as _util
from ..diagbeam.bpm.csdev import Const as _CSBPM
from ..epics import CAThread as _Thread, PV as _PV
from ..thread import RepeaterThread as _Repeat
from .base_class import BaseClass as _BaseClass
from .bpms import BPM, TimingConfig


class BaseOrbit(_BaseClass):
    """."""


class EpicsOrbit(BaseOrbit):
    """Class to deal with orbit acquisition."""

    TIMEOUT_TRIG_ACQ_NEWS = 2 / 5  # [s]

    def __init__(self, acc, prefix="", callback=None):
        """Initialize the instance."""
        super().__init__(acc, prefix=prefix, callback=callback)

        self._mode = 0  # first mode of the list
        self._sofb = None
        self._sync_with_inj = False
        self.ref_orbs = {
            "X": _np.zeros(self._csorb.nr_bpms),
            "Y": _np.zeros(self._csorb.nr_bpms),
        }
        self._load_ref_orbs()
        self.raw_orbs = {"X": [], "Y": []}
        self.raw_sporbs = {"X": [], "Y": [], "Sum": []}
        self.raw_mtorbs = {"X": [], "Y": [], "Sum": []}
        self._lock_raw_orbs = Lock()
        self.smooth_orb = {"X": None, "Y": None}
        self.smooth_sporb = {"X": None, "Y": None, "Sum": None}
        self.smooth_mtorb = {"X": None, "Y": None, "Sum": None}
        self._smooth_npts = 1
        self._smooth_meth = self._csorb.SmoothMeth.Average
        self._spass_mask = [0, 0]
        self._spass_average = 1
        self._acqtrignrsamplespre = 0
        self._acqtrignrsamplespost = 360
        self._acqtrignrshots = 1
        self._multiturnidx = 0
        self._mturndownsample = 1
        self._timevector = None
        self.bpms = [BPM(name, callback) for name in self._csorb.bpm_names]
        self.timing = TimingConfig(acc, callback)
        self.new_orbit = _Event()
        self._new_orbraw_flag = _Event()
        if self.acc == "SI":
            self._sorbraw_pv = _PV(
                "SI-Glob:AP-SOFB:SlowOrbRaw-Mon",
                callback=self._update_sloworb_raw,
                auto_monitor=True,
            )
            self._smooth_npts = 20
        self._timestamp_last_news = 0  # [s] timestamp in epoch
        self._last_num_news = 0
        self._orbit_thread = _Repeat(
            1 / self._csorb.ORBIT_UPDATE_RATE, self._update_orbits, niter=0
        )
        self._orbit_thread.start()
        self._thread_sync = None
        self._update_time_vector()

    @property
    def connected(self):
        """."""
        if self.acc == 'SI' and not self._sorbraw_pv.connected:
            return False
        for bpm in self.bpms:
            if not bpm.connected:
                return False
        if not self.timing.connected:
            return False
        return True

    @property
    def sofb(self):
        """."""
        return self._sofb

    @sofb.setter
    def sofb(self, sofb):
        self._sofb = sofb

    def wait_for_connection(self, timeout=10):
        """."""
        t0_ = _time.time()
        tout = timeout
        if self.acc == 'SI' and not self._sorbraw_pv.wait_for_connection(tout):
            return False
        for bpm in self.bpms:
            tout = timeout - (_time.time() - t0_)
            if tout <= 0 or not bpm.wait_for_connection(tout):
                return False
        tout = timeout - (_time.time() - t0_)
        if tout <= 0 or not self.timing.wait_for_connection(tout):
            return False
        return True

    def shutdown(self):
        """Shutdown threads."""
        self._orbit_thread.resume()
        self._orbit_thread.stop()
        self._orbit_thread.join()

    def get_map2write(self):
        """Get the write methods of the class."""
        dbase = {
            "SOFBMode-Sel": self.set_orbit_mode,
            "SyncWithInjection-Sel": self.set_sync_with_injection,
            "TrigAcqConfig-Cmd": self.acq_config_bpms,
            "TrigAcqCtrl-Sel": self.set_trig_acq_control,
            "TrigAcqChan-Sel": self.set_trig_acq_channel,
            "TrigAcqRepeat-Sel": self.set_trig_acq_repeat,
            "TrigNrSamplesPre-SP": _part(self.set_acq_nrsamples, ispost=False),
            "TrigNrSamplesPost-SP": _part(self.set_acq_nrsamples, ispost=True),
            "RefOrbX-SP": _part(self.set_reforb, "X"),
            "RefOrbY-SP": _part(self.set_reforb, "Y"),
            "SmoothNrPts-SP": self.set_smooth_npts,
            "SmoothMethod-Sel": self.set_smooth_method,
            "SmoothReset-Cmd": self.set_smooth_reset,
            "SPassMaskSplBeg-SP": _part(self.set_spass_mask, beg=True),
            "SPassMaskSplEnd-SP": _part(self.set_spass_mask, beg=False),
            "SPassAvgNrTurns-SP": self.set_spass_average,
            "TrigNrShots-SP": self.set_trig_acq_nrshots,
            "PolyCalibration-Sel": self.set_poly_calibration,
            "SyncBPMs-Cmd": self.sync_bpms,
            "TestDataEnbl-Sel": self.set_test_data_enbl,
        }
        if not self.isring:
            return dbase
        dbase.update(
            {
                "MTurnAcquire-Cmd": self.acquire_mturn_orbit,
                "MTurnIdx-SP": self.set_orbit_multiturn_idx,
                "MTurnDownSample-SP": self.set_mturndownsample,
                "MTurnUseMask-Sel": self.set_mturn_usemask,
                "MTurnMaskSplBeg-SP": _part(self.set_mturnmask, beg=True),
                "MTurnMaskSplEnd-SP": _part(self.set_mturnmask, beg=False),
            }
        )
        return dbase

    @property
    def mode(self):
        """."""
        return self._mode

    @property
    def acqtrignrsamples(self):
        """."""
        return self._acqtrignrsamplespre + self._acqtrignrsamplespost

    @property
    def update_raws(self):
        """."""
        return not self._sync_with_inj or self.timing.injecting

    def is_sloworb(self, mode=None):
        """Check is mode or self._mode is in SlowOrb mode."""
        if mode is None:
            mode = self._mode
        return self.acc == "SI" and mode == self._csorb.SOFBMode.SlowOrb

    def is_multiturn(self, mode=None):
        """Check is mode or self._mode is in MultiTurn mode."""
        if mode is None:
            mode = self._mode
        return self.isring and mode == self._csorb.SOFBMode.MultiTurn

    def is_singlepass(self, mode=None):
        """Check is mode or self._mode is in SinglePass mode."""
        if mode is None:
            mode = self._mode
        return mode == self._csorb.SOFBMode.SinglePass

    def is_trigmode(self, mode=None):
        """Check is mode or self._mode is in any of the Triggered modes."""
        return self.is_singlepass(mode) or self.is_multiturn(mode)

    def get_orbit(self, reset=False, synced=False, timeout=1 / 5):
        """Return the orbit distortion."""
        nrb = self._csorb.nr_bpms
        refx = self.ref_orbs["X"][:nrb]
        refy = self.ref_orbs["Y"][:nrb]
        if reset:
            with self._lock_raw_orbs:
                msg = "DEB: Reseting Orbit."
                _log.debug(msg)
                self._reset_orbs()
                msg = "DEB: Reseted Orbit."
                _log.debug(msg)
            _time.sleep(self._smooth_npts / self._csorb.BPMsFreq)

        if self.is_multiturn():
            orbs = self.smooth_mtorb
            raws = self.raw_mtorbs
            getorb = self._get_orbit_multiturn
        elif self.is_singlepass():
            orbs = self.smooth_sporb
            raws = self.raw_sporbs
            getorb = self._get_orbit_singlepass
        elif self.is_sloworb():
            if synced:
                self.new_orbit.wait(timeout=timeout)
                self.new_orbit.clear()
            orbs = self.smooth_orb
            raws = self.raw_orbs
            getorb = self._get_orbit_online

        for _ in range(3 * self._smooth_npts):
            with self._lock_raw_orbs:
                isempty = orbs["X"] is None or orbs["Y"] is None
                if not isempty and len(raws["X"]) >= self._smooth_npts:
                    orbx, orby = getorb(orbs)
                    break
            msg = "DEB: Trying to get: "
            msg += f'empty={str(isempty):s}, smooth={len(raws["X"]):02d}.'
            _log.debug(msg)
            _time.sleep(1 / self._csorb.BPMsFreq)
        else:
            msg = "ERR: timeout waiting orbit."
            self._update_log(msg)
            _log.error(msg[5:])
            orbx, orby = refx.copy(), refy.copy()
        return _np.hstack([orbx - refx, orby - refy])

    def _get_orbit_online(self, orbs):
        """."""
        return orbs["X"], orbs["Y"]

    def _get_orbit_singlepass(self, orbs):
        """."""
        return orbs["X"], orbs["Y"]

    def _get_orbit_multiturn(self, orbs):
        """."""
        idx = self._multiturnidx
        return orbs["X"][idx, :], orbs["Y"][idx, :]

    def set_smooth_npts(self, num):
        """."""
        self._smooth_npts = num
        self.run_callbacks("SmoothNrPts-RB", num)
        return True

    def set_smooth_method(self, meth):
        """."""
        self._smooth_meth = meth
        self.run_callbacks("SmoothMethod-Sts", meth)
        return True

    def set_spass_mask(self, val, beg=True):
        """."""
        val = int(val) if val > 0 else 0
        other_mask = self._spass_mask[1 if beg else 0]
        maxsz = self.bpms[0].tbtrate - other_mask - 2
        val = val if val < maxsz else maxsz
        self._spass_mask[0 if beg else 1] = val
        name = "Beg" if beg else "End"
        self.run_callbacks("SPassMaskSpl" + name + "-RB", val)
        return True

    def set_mturn_usemask(self, val):
        """."""
        value = _CSBPM.DsblEnbl.enabled
        if val == self._csorb.DsblEnbl.Dsbl:
            value = _CSBPM.DsblEnbl.disabled

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.tbt_mask_enbl = value

        self.run_callbacks("MTurnUseMask-Sts", val)
        return True

    def set_mturnmask(self, val, beg=True):
        """."""
        val = int(val) if val > 0 else 0
        bpms = self._get_used_bpms()
        omsk = bpms[0].tbt_mask_begin if not beg else bpms[0].tbt_mask_end
        omsk = omsk or 0
        maxsz = bpms[0].tbtrate - omsk - 2
        val = val if val < maxsz else maxsz

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            if beg:
                bpm.tbt_mask_begin = val
            else:
                bpm.tbt_mask_end = val

        name = "Beg" if beg else "End"
        self.run_callbacks("MTurnMaskSpl" + name + "-RB", val)
        return True

    def set_spass_average(self, val):
        """."""
        val = int(val) if val > 1 else 1
        with self._lock_raw_orbs:
            self._spass_average = val
            self._reset_orbs()
        self.run_callbacks("SPassAvgNrTurns-RB", val)
        self._prepare_mode()
        return True

    def set_smooth_reset(self, _):
        """."""
        with self._lock_raw_orbs:
            self._reset_orbs()
        return True

    def set_reforb(self, plane, orb):
        """."""
        msg = "Setting New Reference Orbit."
        self._update_log(msg)
        _log.info(msg)
        orb = _np.array(orb, dtype=float)
        nrb = self._csorb.nr_bpms
        if orb.size % self._csorb.nr_bpms:
            msg = "ERR: Wrong RefOrb Size."
            self._update_log(msg)
            _log.error(msg[5:])
            self.run_callbacks(
                "RefOrb" + plane + "-SP", self.ref_orbs[plane][:nrb]
            )
        elif orb.size < nrb:
            msg = "WARN: Orb Size is too small. Replicating..."
            self._update_log(msg)
            _log.error(msg[6:])
            nrep = int(nrb // orb.size) + 1
            orb2 = _np.tile(orb, nrep)
            orb = orb2[:nrb]
        self.ref_orbs[plane] = orb
        self._save_ref_orbits()
        with self._lock_raw_orbs:
            self._reset_orbs()
        self.run_callbacks("RefOrb" + plane + "-RB", orb[:nrb])
        return True

    def set_orbit_mode(self, value):
        """."""
        omode = self._mode
        with self._lock_raw_orbs:
            self._mode = value
            self._reset_orbs()
        if self.acc == 'SI':
            self.run_callbacks("SOFBMode-Sts", value)
        self._prepare_mode(oldmode=omode)
        return True

    def set_sync_with_injection(self, boo):
        """."""
        self._sync_with_inj = bool(boo)
        self.run_callbacks("SyncWithInjection-Sts", bool(boo))
        return True

    def set_test_data_enbl(self, val, is_thread=False):
        """."""
        if not is_thread:
            self._LQTHREAD.put(
                (self.set_test_data_enbl, (val,), {"is_thread": True})
            )
            return True

        value = _CSBPM.DsblEnbl.enabled
        if val == self._csorb.DsblEnbl.Dsbl:
            value = _CSBPM.DsblEnbl.disabled

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.test_data_enbl = value
        self.run_callbacks("TestDataEnbl-Sts", val)

    def _prepare_mode(self, oldmode=None):
        """."""
        oldmode = self._mode if oldmode is None else oldmode
        self.set_trig_acq_control(self._csorb.TrigAcqCtrl.Stop)

        if not self.is_trigmode():
            self.acq_config_bpms()
            return True

        if self.is_singlepass():
            chan = self._csorb.TrigAcqChan.ADCSwp
            rep = self._csorb.TrigAcqRepeat.Repetitive
            points = self._spass_average * self.bpms[0].tbtrate
        elif self.is_multiturn():
            chan = self._csorb.TrigAcqChan.TbT
            rep = self._csorb.TrigAcqRepeat.Repetitive
            points = self._mturndownsample

        if self._mode != oldmode:
            self.run_callbacks("TrigAcqChan-Sel", chan)
            self.set_trig_acq_channel(chan)
            self.run_callbacks("TrigAcqRepeat-Sel", rep)
            self.set_trig_acq_repeat(rep)
            if self.acqtrignrsamples < points:
                pts = points - self._acqtrignrsamplespre
                self.run_callbacks("TrigNrSamplesPost-SP", pts)
                self.set_acq_nrsamples(pts, ispost=True)
        self._update_time_vector()
        self.acq_config_bpms()

        self.set_trig_acq_control(self._csorb.TrigAcqCtrl.Start)
        return True

    def set_orbit_multiturn_idx(self, value):
        """."""
        maxidx = self.acqtrignrsamples // self._mturndownsample
        maxidx *= self._acqtrignrshots
        if value >= maxidx:
            value = maxidx - 1
            msg = "WARN: MTurnIdx is too large. Redefining..."
            self._update_log(msg)
            _log.warning(msg[6:])
        with self._lock_raw_orbs:
            self._multiturnidx = int(value)
        self.run_callbacks("MTurnIdx-RB", self._multiturnidx)
        self.run_callbacks(
            "MTurnIdxTime-Mon", self._timevector[self._multiturnidx]
        )
        self._update_multiturn_orbit_pvs()
        return True

    def acq_config_bpms(self, *args):
        """."""
        _ = args
        msg = "Configuring BPMs..."
        self._update_log(msg)
        _log.info(msg)

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            if self.is_multiturn():
                bpm.switching_mode = _CSBPM.SwModes.direct
                bpm.configure()
                self.timing.configure()
            elif self.is_singlepass():
                bpm.switching_mode = _CSBPM.SwModes.direct
                bpm.configure()
                self.timing.configure()
            elif self.is_sloworb():
                bpm.switching_mode = _CSBPM.SwModes.switching
        self.sync_bpms(None)
        msg = "Done configuring BPMs!"
        self._update_log(msg)
        _log.info(msg)
        return True

    def sync_bpms(self, *args):
        """Synchronize BPMs."""
        _ = args

        msg = "Received sync BPMs command."
        self._update_log(msg)
        _log.info(msg)

        if self._thread_sync is not None and self._thread_sync.is_alive():
            msg = "WARN: Previous sync. of BPMs still running."
            self._update_log(msg)
            _log.warning(msg)
            return False

        self._thread_sync = _Thread(target=self._synchronize_bpms, daemon=True)
        self._thread_sync.start()
        return True

    def _synchronize_bpms(self):
        msg = "Syncing BPMs."
        self._update_log(msg)
        _log.info(msg)
        for bpm in self._get_used_bpms():
            # NOTE: Switching sync must always be enabled
            bpm.sw_sync_enbl = _CSBPM.DsblEnbl.enabled

            bpm.tbt_sync_enbl = _CSBPM.DsblEnbl.enabled
            bpm.fofb_sync_enbl = _CSBPM.DsblEnbl.enabled
            bpm.facq_sync_enbl = _CSBPM.DsblEnbl.enabled
            bpm.monit_sync_enbl = _CSBPM.DsblEnbl.enabled
        _time.sleep(0.5)
        for bpm in self._get_used_bpms():
            bpm.tbt_sync_enbl = _CSBPM.DsblEnbl.disabled
            bpm.fofb_sync_enbl = _CSBPM.DsblEnbl.disabled
            bpm.monit_sync_enbl = _CSBPM.DsblEnbl.disabled
            bpm.facq_sync_enbl = _CSBPM.DsblEnbl.disabled

        if self.acc == "SI" and self.sofb.fofb.connected:
            _time.sleep(0.2)
            msg = "Syncing FOFB Net..."
            self._update_log(msg)
            _log.info(msg)
            self.sofb.fofb.cmd_fofbctrl_syncnet()

        msg = "Syncing BPMs is done!"
        self._update_log(msg)
        _log.info(msg)

    def set_trig_acq_control(self, value):
        """."""
        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.ctrl = value
        self.run_callbacks("TrigAcqCtrl-Sts", value)
        return True

    def set_trig_acq_channel(self, value):
        """."""
        try:
            val = self._csorb.TrigAcqChan._fields[value]
            val = _CSBPM.AcqChan._fields.index(val)
        except (IndexError, ValueError):
            return False

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.acq_type = val

        self.run_callbacks("TrigAcqChan-Sts", value)
        self._update_time_vector(channel=val)
        return True

    def set_trig_acq_repeat(self, value):
        """."""
        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.acq_repeat = value
        self.run_callbacks("TrigAcqRepeat-Sts", value)
        return True

    def set_acq_nrsamples(self, val, ispost=True):
        """."""
        val = int(val) if val > 0 else 0
        val = val if val < 20000 else 20000
        suf = "post" if ispost else "pre"
        oth = "post" if not ispost else "pre"
        if getattr(self, "_acqtrignrsamples" + oth) == 0 and val == 0:
            self.run_callbacks(
                "TrigNrSamples" + suf.title() + "-SP",
                getattr(self, "_acqtrignrsamples" + suf),
            )
        with self._lock_raw_orbs:
            mask = self._get_mask()
            for i, bpm in enumerate(self.bpms):
                bpm.put_enable = mask[i]
                setattr(bpm, "nrsamples" + suf, val)
            self._reset_orbs()
            setattr(self, "_acqtrignrsamples" + suf, val)
        self.run_callbacks("TrigNrSamples" + suf.title() + "-RB", val)
        self._update_time_vector()
        return True

    def set_trig_acq_nrshots(self, val):
        """."""
        val = int(val) if val > 1 else 1
        val = val if val < 1000 else 1000
        with self._lock_raw_orbs:
            mask = self._get_mask()
            for i, bpm in enumerate(self.bpms):
                bpm.put_enable = mask[i]
                bpm.nrshots = val
            self.timing.nrpulses = val
            self._reset_orbs()
            self._acqtrignrshots = val
        self.run_callbacks("TrigNrShots-RB", val)
        self._update_time_vector()
        return True

    def set_poly_calibration(self, val):
        """."""
        value = _CSBPM.DsblEnbl.enabled
        if val == self._csorb.DsblEnbl.Dsbl:
            value = _CSBPM.DsblEnbl.disabled
        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.polycal = value
        self.run_callbacks("PolyCalibration-Sts", val)
        return True

    def set_mturndownsample(self, val):
        """."""
        val = int(val) if val > 1 else 1
        val = val if val < 1000 else 1000
        with self._lock_raw_orbs:
            self._mturndownsample = val
            self._reset_orbs()
        self.run_callbacks("MTurnDownSample-RB", val)
        self._prepare_mode()
        return True

    def acquire_mturn_orbit(self, _):
        """Acquire Multiturn data from BPMs."""
        self._update_multiturn_orbits(force_update=True)
        return True

    def _update_time_vector(self, delay=None, duration=None, channel=None):
        """."""
        if not self.isring:
            return
        dly = (delay or self.timing.totaldelay or 0.0) * 1e-6  # from us to s
        dur = (duration or self.timing.duration or 0.0) * 1e-6  # from us to s
        channel = channel or self.bpms[0].acq_type or 0

        # revolution period in s
        if channel == _CSBPM.AcqChan.FAcq:
            dtime = self.bpms[0].facqperiod
        elif channel == _CSBPM.AcqChan.FOFB:
            dtime = self.bpms[0].fofbperiod
        else:
            dtime = self.bpms[0].tbtperiod
        mult = self._mturndownsample
        dtime *= mult
        nrptpst = self.acqtrignrsamples // mult
        offset = self._acqtrignrsamplespre / mult
        nrst = self._acqtrignrshots
        shots = _np.arange(nrst)
        pts = _np.arange(nrptpst, dtype=float) + (0.5 - offset)
        vect = dly + dur / nrst * shots[:, None] + dtime * pts[None, :]
        self._timevector = vect.ravel()
        self.run_callbacks("MTurnTime-Mon", self._timevector)
        self.set_orbit_multiturn_idx(self._multiturnidx)

    def _load_ref_orbs(self):
        """."""
        if not _os.path.isfile(self._csorb.ref_orb_fname):
            return
        self.ref_orbs["X"], self.ref_orbs["Y"] = _np.loadtxt(
            self._csorb.ref_orb_fname, unpack=True
        )
        self.run_callbacks("RefOrbX-RB", self.ref_orbs["X"].copy())
        self.run_callbacks("RefOrbY-RB", self.ref_orbs["Y"].copy())

    def _save_ref_orbits(self):
        """."""
        refx = self.ref_orbs["X"]
        refy = self.ref_orbs["Y"]
        if refx.size < refy.size:
            ref = _np.zeros(refy.shape, dtype=float)
            ref[: refx.size] = refx
            refx = ref
        elif refy.size < refx.size:
            ref = _np.zeros(refx.shape, dtype=float)
            ref[: refy.size] = refy
            refy = ref
        orbs = _np.array([refx, refy]).T
        try:
            path = _os.path.split(self._csorb.ref_orb_fname)[0]
            _os.makedirs(path, exist_ok=True)
            _np.savetxt(self._csorb.ref_orb_fname, orbs)
        except FileNotFoundError:
            msg = "WARN: Could not save reference orbit in file."
            self._update_log(msg)
            _log.warning(msg[6:])

    def _reset_orbs(self):
        """."""
        raw = self.raw_orbs
        smt = self.smooth_orb
        raw["X"], raw["Y"] = [], []
        smt["X"], smt["Y"] = None, None
        raw = self.raw_sporbs
        smt = self.smooth_sporb
        raw["X"], raw["Y"], raw["Sum"] = [], [], []
        smt["X"], smt["Y"], smt["Sum"] = None, None, None
        raw = self.raw_mtorbs
        smt = self.smooth_mtorb
        raw["X"], raw["Y"], raw["Sum"] = [], [], []
        smt["X"], smt["Y"], smt["Sum"] = None, None, None
        self.run_callbacks("BufferCount-Mon", 0)

    def _update_orbits(self):
        """."""
        try:
            count = 0
            if self.is_multiturn():
                self._update_multiturn_orbits(force_update=False)
                count = len(self.raw_mtorbs["X"])
            elif self.is_singlepass():
                self._update_singlepass_orbits()
                count = len(self.raw_sporbs["X"])
            elif self.is_sloworb():
                self._update_online_orbits()
                count = len(self.raw_orbs["X"])
            self.run_callbacks("BufferCount-Mon", count)
        except Exception as err:
            self._update_log("ERR: " + str(err))
            _log.error(_traceback.format_exc())

    def _update_online_orbits(self):
        """."""
        timeout = 1000 / 1000
        if not self._new_orbraw_flag.wait(timeout=timeout):
            msg = "ERR: Raw orbit did not update."
            self._update_log(msg)
            _log.error(msg[5:])
            return
        else:
            self._new_orbraw_flag.clear()

        orb = self._sorbraw_pv.value
        posx, posy = orb[: self._csorb.nr_bpms], orb[self._csorb.nr_bpms :]
        nanx = _np.isnan(posx)
        nany = _np.isnan(posy)
        posx[nanx] = self.ref_orbs["X"][nanx]
        posy[nany] = self.ref_orbs["Y"][nany]
        orbs = {"X": posx, "Y": posy}

        for plane in ("X", "Y"):
            with self._lock_raw_orbs:
                raws = self.raw_orbs
                raws[plane].append(orbs[plane])
                raws[plane] = raws[plane][-self._smooth_npts:]
                if not raws[plane]:
                    return
                if self._smooth_meth == self._csorb.SmoothMeth.Average:
                    orb = _np.mean(raws[plane], axis=0)
                else:
                    orb = _np.median(raws[plane], axis=0)
                self.smooth_orb[plane] = orb
        self.new_orbit.set()

        for plane in ("X", "Y"):
            orb = self.smooth_orb[plane]
            if orb is None:
                return
            dorb = orb - self.ref_orbs[plane]
            self.run_callbacks(f"SlowOrb{plane:s}-Mon", _np.array(orb))
            self.run_callbacks(f"DeltaOrb{plane:s}Avg-Mon", _bn.nanmean(dorb))
            self.run_callbacks(f"DeltaOrb{plane:s}Std-Mon", _bn.nanstd(dorb))
            self.run_callbacks(f"DeltaOrb{plane:s}Min-Mon", _bn.nanmin(dorb))
            self.run_callbacks(f"DeltaOrb{plane:s}Max-Mon", _bn.nanmax(dorb))

    def _update_sloworb_raw(self, pvname, value, **kwrgs):
        _ = pvname, value, kwrgs
        self._new_orbraw_flag.set()

    def _update_multiturn_orbits(self, force_update=True):
        """."""
        if not force_update and not self.update_raws:
            return

        news = sum([bpm.has_news for bpm in self.bpms])
        if news > self._last_num_news:
            self._last_num_news = news
            self._timestamp_last_news = _time.time()

        hasnews = news == len(self.bpms)
        dtime = _time.time() - self._timestamp_last_news
        timeout = dtime >= self.TIMEOUT_TRIG_ACQ_NEWS
        if not hasnews and not timeout and len(self.raw_mtorbs["X"]):
            return

        orbs = {"X": [], "Y": [], "Sum": []}
        with self._lock_raw_orbs:  # Lock needed here to ensure consistency
            leng = len(self.raw_mtorbs["X"])
            samp = self.acqtrignrsamples
            down = self._mturndownsample
            samp -= samp % down
            if samp < 1:
                msg = "ERR: Actual nr_samples in MTurn orb calc. is < 1."
                self._update_log(msg)
                _log.error(msg[5:])
                return
            samp *= self._acqtrignrshots
            orbsz = self._csorb.nr_bpms
            nr_pts = self._smooth_npts
            isdiff = False  # only update orbit if some BPM updated
            for i, bpm in enumerate(self.bpms):
                if not leng or bpm.has_news:
                    posx = self._get_pos(
                        bpm.mtposx, self.ref_orbs["X"][i], samp
                    )
                    posy = self._get_pos(
                        bpm.mtposy, self.ref_orbs["Y"][i], samp
                    )
                    psum = self._get_pos(bpm.mtsum, 0, samp)
                    thisdiff = not leng or not _np.array_equal(
                        posx, self.raw_mtorbs["X"][-1][:, i]
                    )
                    # if it got here, then for sure the new data will be used
                    # and we can reset the flag:
                    if thisdiff:
                        bpm.has_news = False
                    isdiff |= thisdiff
                else:
                    posx = self.raw_mtorbs["X"][-1][:, i].copy()
                    posy = self.raw_mtorbs["Y"][-1][:, i].copy()
                    psum = self.raw_mtorbs["Sum"][-1][:, i].copy()
                orbs["X"].append(posx)
                orbs["Y"].append(posy)
                orbs["Sum"].append(psum)

            if not isdiff:
                return
            self._timestamp_last_news = _time.time()
            self._last_num_news = 0

            for pln, raw in self.raw_mtorbs.items():
                norb = _np.array(orbs[pln], dtype=float)  # bpms x turns
                norb = norb.T.reshape(-1, orbsz)  # turns/rz x rz*bpms
                raw.append(norb)
                del raw[:-nr_pts]
                if not raw:
                    return
                if self._smooth_meth == self._csorb.SmoothMeth.Average:
                    orb = _np.mean(raw, axis=0)
                else:
                    orb = _np.median(raw, axis=0)
                if down > 1:
                    orb = _np.mean(orb.reshape(-1, down, orbsz), axis=1)
                self.smooth_mtorb[pln] = orb
                orbs[pln] = orb
        self._update_multiturn_orbit_pvs()

    def _update_multiturn_orbit_pvs(self):
        for pln, orb in self.smooth_mtorb.items():
            if orb is None:
                continue
            idx = min(self._multiturnidx, orb.shape[0])
            name = ("Orb" if pln != "Sum" else "") + pln
            self.run_callbacks("MTurn" + name + "-Mon", orb.ravel())
            self.run_callbacks("MTurnIdx" + name + "-Mon", orb[idx, :].ravel())
            if pln == "Sum":
                continue
            nrb = self._csorb.nr_bpms
            dorb = orb[idx, :].ravel() - self.ref_orbs[pln][:nrb]
            self.run_callbacks(f"DeltaOrb{pln:s}Avg-Mon", _bn.nanmean(dorb))
            self.run_callbacks(f"DeltaOrb{pln:s}Std-Mon", _bn.nanstd(dorb))
            self.run_callbacks(f"DeltaOrb{pln:s}Min-Mon", _bn.nanmin(dorb))
            self.run_callbacks(f"DeltaOrb{pln:s}Max-Mon", _bn.nanmax(dorb))

    @staticmethod
    def _get_pos(pos, ref, samp):
        if pos is None:
            posi = _np.full(samp, ref)
        elif pos.size < samp:
            posi = _np.full(samp, ref)
            posi[: pos.size] = pos
        else:
            posi = pos[:samp]
        return posi

    def _update_singlepass_orbits(self):
        """."""
        if not self.update_raws:
            return

        news = sum([bpm.has_news for bpm in self.bpms])
        if news > self._last_num_news:
            self._last_num_news = news
            self._timestamp_last_news = _time.time()

        hasnews = news == len(self.bpms)
        dtime = _time.time() - self._timestamp_last_news
        timeout = dtime >= self.TIMEOUT_TRIG_ACQ_NEWS
        if not hasnews and not timeout and len(self.raw_sporbs["X"]):
            return

        orbs = {"X": [], "Y": [], "Sum": []}
        down = self._spass_average
        with self._lock_raw_orbs:  # Lock needed here to ensure consistency
            leng = len(self.raw_sporbs["X"])
            dic = {
                "maskbeg": self._spass_mask[0],
                "maskend": self._spass_mask[1],
                "nturns": down,
            }
            nr_pts = self._smooth_npts
            isdiff = False  # only update orbit if some BPM updated
            for i, bpm in enumerate(self.bpms):
                if not leng or bpm.has_news:
                    dic.update(
                        {
                            "refx": self.ref_orbs["X"][i],
                            "refy": self.ref_orbs["Y"][i],
                        }
                    )
                    orbx, orby, summ = bpm.calc_sp_multiturn_pos(**dic)
                    thisdiff = not leng or not _np.array_equal(
                        orbx, self.raw_sporbs["X"][-1][i]
                    )
                    # if it got here, then for sure the new data will be used
                    # and we can reset the flag:
                    if thisdiff:
                        bpm.has_news = False
                    isdiff |= thisdiff
                else:
                    orbx = self.raw_sporbs["X"][-1][i].copy()
                    orby = self.raw_sporbs["Y"][-1][i].copy()
                    summ = self.raw_sporbs["Sum"][-1][i].copy()
                orbs["X"].append(orbx)
                orbs["Y"].append(orby)
                orbs["Sum"].append(summ)

            if not isdiff:
                return
            self._timestamp_last_news = _time.time()
            self._last_num_news = 0

            for pln, raw in self.raw_sporbs.items():
                norb = _np.array(orbs[pln], dtype=float).T  # turns x bpms
                norb = norb.reshape(-1)
                raw.append(norb)
                del raw[:-nr_pts]
                if not raw:
                    return
                if self._smooth_meth == self._csorb.SmoothMeth.Average:
                    orb = _np.mean(raw, axis=0)
                else:
                    orb = _np.median(raw, axis=0)
                if down > 1:
                    orb = _np.mean(orb.reshape(down, -1), axis=0)
                self.smooth_sporb[pln] = orb

        for pln, orb in self.smooth_sporb.items():
            name = ("Orb" if pln != "Sum" else "") + pln
            self.run_callbacks("SPass" + name + "-Mon", orb)

    def _update_status(self):
        """."""
        status = 0b11111

        if self.is_trigmode():
            tim_conn = self.timing.connected
            tim_conf = self.timing.is_ok
        else:
            tim_conn = True
            tim_conf = True
        status = _util.update_bit(status, bit_pos=0, bit_val=not tim_conn)
        status = _util.update_bit(status, bit_pos=1, bit_val=not tim_conf)

        bpms = self._get_used_bpms()
        bpm_conn = all(bpm.connected for bpm in bpms)

        status = _util.update_bit(v=status, bit_pos=2, bit_val=not bpm_conn)

        isok = True
        if self.is_trigmode():
            isok = all(map(lambda x: x.is_ok, bpms))
        elif self.is_sloworb():
            isok = all(
                map(
                    lambda x: x.switching_mode == _CSBPM.SwModes.switching,
                    bpms,
                )
            )
        status = _util.update_bit(v=status, bit_pos=3, bit_val=not isok)

        # Check if test data is disabled
        isok = all(
            map(lambda x: x.test_data_enbl == _CSBPM.DsblEnbl.disabled, bpms)
        )
        status = _util.update_bit(v=status, bit_pos=4, bit_val=not isok)

        # Check if switching sync is enabled
        isok = all(
            map(lambda x: x.sw_sync_enbl == _CSBPM.DsblEnbl.enabled, bpms)
        )
        status = _util.update_bit(v=status, bit_pos=5, bit_val=not isok)

        orb_conn = self._sorbraw_pv.connected if self.acc == "SI" else True
        status = _util.update_bit(v=status, bit_pos=6, bit_val=not orb_conn)

        self._status = status
        self.run_callbacks("OrbStatus-Mon", status)
        self._update_bpmoffsets()

    def _update_bpmoffsets(self):
        """."""
        nrb = self._csorb.nr_bpms
        orbx = _np.zeros(nrb, dtype=float)
        orby = orbx.copy()
        for i, bpm in enumerate(self.bpms):
            orbx[i::nrb] = bpm.offsetx or 0.0
            orby[i::nrb] = bpm.offsety or 0.0
        self.run_callbacks("BPMOffsetX-Mon", orbx)
        self.run_callbacks("BPMOffsetY-Mon", orby)

    def _get_mask(self):
        mask = _np.ones(self._csorb.nr_bpms, dtype=bool)
        if self.sofb is not None and self.sofb.matrix is not None:
            mask = self.sofb.matrix.bpm_enbllist
            mask = mask[: mask.size // 2] & mask[mask.size // 2 :]
            mask = mask[: len(self.bpms)]
        return mask

    def _get_used_bpms(self):
        bpms = []
        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            if mask[i]:
                bpms.append(bpm)
        return bpms
